# -*- coding: utf-8 -*-
"""Storj tests."""

import logging
import unittest


from abc import ABCMeta

from . import config
from . import log_config


config.load_configuration()
log_config.load_configuration()


class Basic(object):
    """Basic functionality to enhance test cases.

    Attributes:
        configuration ():
        logger (): logger.
    """
    __metaclass__ = ABCMeta

    def setup_configuration(self):
        """
        Setup test configuration.
        It will also load (once) the test configuration.
        """
        logging.getLogger(
            '%s.%s' % (__name__, 'Basic')).info('setup_configuration()')

        self.configuration = config.get()

    def setup_logger(self):
        """
        Setup test logger.
        It will also load (once) the test logging configuration.
        """
        logging.getLogger('%s.%s' % (__name__, 'Basic')).info('setup_logger()')

        self.logger = logging.getLogger(
            '%s.%s' % (__name__, self.__class__.__name__))


class AbstractTestCase(unittest.TestCase, Basic):
    """Base test case."""
    __metaclass__ = ABCMeta

    __slots__ = ('configuration', 'logger')

    def __init__(self, methodName):
        """Initializes a BaseTestCase instance.

        Args:
            methodName (str): the test method to be executed.
        """
        super(AbstractTestCase, self).__init__(methodName)

        self.setup_logger()
        self.setup_configuration()

    def setUp(self):
        """Setup test resources."""
        self.logger.info('setUp()')

    def tearDown(self):
        """Tear down test resources."""
        self.logger.info('tearDown()')



# -*- coding: utf-8 -*-
"""Storj HTTP module."""

import os

import logging
import json
import requests
import time

from base64 import b64encode
from binascii import b2a_hex

from ecdsa import SigningKey, SECP256k1, VerifyingKey
from ecdsa.util import sigencode_der
from hashlib import sha256
from io import BytesIO
from six.moves.urllib.parse import urlencode, urljoin

try:
    from json.decoder import JSONDecodeError
except ImportError:
    # Python 2
    JSONDecodeError = ValueError

from api import ecdsa_to_hex
from exception import BridgeError, ClientError

from . import web_socket
from . import model


__logger = logging.getLogger(__name__)


def handle_nonhttp_errors(func):
    """Handle non-HTTP errors."""
    def decorator(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.exceptions.RequestException as e:
            __logger.error('Failed to contact the bridge: %s', e)
            raise ClientError(message='Failed to contact the bridge')
    return decorator


class Client(object):
    """
    Attributes:
        api_url (str): the Storj API endpoint.
        session ():
        email (str): user email address.
        password (str): user password.
        private_key ():
        public_key ():
        public_key_hex ():
    """

    logger = logging.getLogger('%s.Client' % __name__)

    def __init__(self, email, password, do_hashing=True):
        self.api_url = 'https://api.storj.io/'
        self.session = requests.Session()
        self.email = email
        self.do_hashing = do_hashing
        self.password = password
        self.private_key = None
        self.public_key = None
        self.public_key_hex = None

    @property
    def password(self):
        """(str): user password"""
        return self._password

    @password.setter
    def password(self, value):
        if self.do_hashing:
            self._password = sha256(value.encode('ascii')).hexdigest()
        else:
            self._password = value

    def authenticate(self, ecdsa_private_key=None):
        self.logger.debug('authenticate')

        if isinstance(ecdsa_private_key, SigningKey):
            self.private_key = ecdsa_private_key
            self.public_key = self.private_key.get_verifying_key()
            self.public_key_hex = ecdsa_to_hex(self.public_key.to_string())

    def _add_basic_auth(self, request_kwargs):
        self.logger.debug('using basic auth')

        request_kwargs['headers'].update({
            'Authorization': b'Basic ' + b64encode(
                ('%s:%s' % (self.email, self.password)).encode('ascii')
            ),
        })

    def _add_ecdsa_signature(self, request_kwargs):

        method = request_kwargs.get('method', 'GET')

        if method in ('GET', 'DELETE'):
            request_kwargs.setdefault('params', {})
            request_kwargs['params']['__nonce'] = int(time.time())
            data = urlencode(request_kwargs['params'])
        else:
            request_kwargs.setdefault('json', {})
            request_kwargs['json']['__nonce'] = int(time.time())
            data = json.dumps(request_kwargs['json'])

        contract = '\n'.join(
            (method, request_kwargs['path'], data)).encode('utf-8')

        signature_bytes = self.private_key.sign(
            contract, sigencode=sigencode_der, hashfunc=sha256)

        signature = b2a_hex(signature_bytes).decode('ascii')

        request_kwargs['headers'].update(
            {
                'x-signature': signature,
                'x-pubkey': ecdsa_to_hex(self.public_key.to_string()),
            })

    def _prepare_request(self, **kwargs):
        """Prepares a HTTP request.
        Args:
            kwargs (dict): keyword arguments for the authentication function
                (``_add_ecdsa_signature()`` or ``_add_basic_auth()``) and
                :py:class:`requests.Request` class.
        Raises:
            AssertionError: in case ``kwargs['path']`` doesn't start with ``/``.
        """

        kwargs.setdefault('headers', {})

        # Add appropriate authentication headers
        if isinstance(self.private_key, SigningKey):
            self._add_ecdsa_signature(kwargs)
        elif self.email and self.password:
            self._add_basic_auth(kwargs)

        # Generate URL from path
        path = kwargs.pop('path')
        assert path.startswith('/')
        kwargs['url'] = urljoin(self.api_url, path)

        return requests.Request(**kwargs).prepare()

    def _request(self, **kwargs):
        """Perform HTTP request.
        Args:
            kwargs (dict): keyword arguments.
        Raises:
            :py:class:`BridgeError`: in case::
                - internal server error
                - error attribute is present in the JSON response
                - HTTP response JSON decoding failed
        """

        response = self.session.send(self._prepare_request(**kwargs))
        self.logger.debug('_request response %s', response.text)

        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            self.logger.error(e)
            self.logger.debug('response.text=%s', response.text)
            raise BridgeError(response.text)

        # Raise any errors as exceptions
        try:
            if response.text != '':
                response_json = response.json()
            else:
                return {}

            if 'error' in response_json:
                raise BridgeError(response_json['error'])

            return response_json

        except JSONDecodeError as e:
            self.logger.error(e)
            self.logger.error('_request body %s', response.text)
            raise BridgeError('Could not decode response.')

    @handle_nonhttp_errors
    def bucket_create(self, name, storage=None, transfer=None):
        """Create storage bucket.
        See `API buckets: POST /buckets
        <https://storj.github.io/bridge/#!/buckets/post_buckets>`_
        Args:
            name (str): name.
            storage (int): storage limit (in GB).
            transfer (int): transfer limit (in GB).
        Returns:
            (:py:class:`model.Bucket`): bucket.
        """
        self.logger.info('bucket_create(%s, %s, %s)', name, storage, transfer)

        data = {'name': name}
        if storage:
            data['storage'] = storage
        if transfer:
            data['transfer'] = transfer

        return model.Bucket(**self._request(method='POST', path='/buckets', json=data))

    @handle_nonhttp_errors
    def bucket_delete(self, bucket_id):
        """Destroy a storage bucket.
        See `API buckets: DELETE /buckets/{id}
        <https://storj.github.io/bridge/#!/buckets/delete_buckets_id>`_
        Args:
            bucket_id (string): unique identifier.
        """
        self.logger.info('bucket_delete(%s)', bucket_id)
        self._request(method='DELETE', path='/buckets/%s' % bucket_id)

    @handle_nonhttp_errors
    def bucket_files(self, bucket_id):
        """List all the file metadata stored in the bucket.
        See `API buckets: GET /buckets/{id}/files
        <https://storj.github.io/bridge/#!/buckets/get_buckets_id_files>`_
        Args:
            bucket_id (string): unique identifier.
        Returns:
            (dict): to be changed to model in the future.
        """
        self.logger.info('bucket_files(%s)', bucket_id)

        return self._request(
            method='GET',
            path='/buckets/%s/files/' % (bucket_id), )

    @handle_nonhttp_errors
    def bucket_get(self, bucket_id):
        """Return the bucket object.
        See `API buckets: GET /buckets
        <https://storj.github.io/bridge/#!/buckets/get_buckets_id>`_
        Args:
            bucket_id (str): bucket unique identifier.
        Returns:
            (:py:class:`model.Bucket`): bucket.
        """
        self.logger.info('bucket_get(%s)', bucket_id)

        try:
            return model.Bucket(**self._request(
                method='GET',
                path='/buckets/%s' % bucket_id))

        except requests.HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                return None
            else:
                self.logger.error('bucket_get() error=%s', e)
                raise BridgeError()

    @handle_nonhttp_errors
    def bucket_list(self):
        """List all of the buckets belonging to the user.
        See `API buckets: GET /buckets
        <https://storj.github.io/bridge/#!/buckets/get_buckets>`_
        Returns:
            (generator[:py:class:`model.Bucket`]): buckets.
        """
        self.logger.info('bucket_list()')

        response = self._request(method='GET', path='/buckets')

        if response is not None:
            for element in response:
                yield model.Bucket(**element)
        else:
            raise StopIteration

    @handle_nonhttp_errors
    def bucket_set_keys(self, bucket_id, bucket_name, keys):
        """Update the bucket with the given public keys.
        See `API buckets: PATCH /buckets/{bucket_id}
        <https://storj.github.io/bridge/#!/buckets/patch_buckets_id>`_
        Args:
            bucket_id (str): bucket unique identifier.
            bucket_name (str): bucket name.
            keys (list[str]): public keys.
        Returns:
            (:py:class:`storj.model.Bucket`): updated bucket information.
        """
        self.logger.info('bucket_set_keys(%s, %s)', bucket_name, keys)

        return model.Bucket(**self._request(
            method='PATCH',
            path='/buckets/%s' % bucket_id,
            json={
                'name': bucket_name,
                'pubkeys': keys}))

    @handle_nonhttp_errors
    def bucket_set_mirrors(self, bucket_id, file_id, redundancy):
        """Establishes a series of mirrors for the given file.
        See `API buckets: POST /buckets/{id}/mirrors
        <https://storj.github.io/bridge/#!/buckets/post_buckets_id_mirrors>`_
        Args:
            bucket_id (str): bucket unique identifier.
            file_id (str): file unique identitifer.
            redundancy (int): number of replicas.
        Returns:
            (:py:class:`storj.model.Mirror`): the mirror settings.
        """
        self.logger.info('bucket_set_mirrors(%s, %s, %s)', bucket_id, file_id, redundancy)

        return model.Mirror(**self._request(
            method='POST',
            path='/buckets/%s/mirrors' % bucket_id,
            json={
                'file': file_id,
                'redundancy': redundancy
            }))

    @handle_nonhttp_errors
    def contact_list(self, page=1, address=None, protocol=None, user_agent=None, connected=None):
        """Lists contacts.
        See `API contacts: GET /contacts
        <https://storj.github.io/bridge/#!/contacts/get_contacts>`_
        Args:
            page (str): pagination indicator.
            address (str): hostname or IP address.
            protocol (str): SemVer protocol tag.
            user_agent (str): Storj user agent string for farming client.
            connected (bool): filter results by connection status.
        Returns:
            (list[:py:class:`storj.model.Contact`]): list of contacts.
        """
        self.logger.info('contacts_list()')

        response = self._request(
            method='GET',
            path='/contacts')

        if response is not None:
            for kwargs in response:
                yield model.Contact(**kwargs)
        else:
            raise StopIteration

    @handle_nonhttp_errors
    def contact_lookup(self, node_id):
        """Lookup for contact information of a node.
        See `API contacts: GET /contacts/{nodeID}
        <https://storj.github.io/bridge/#!/contacts/get_contacts_nodeID>`_
        Args:
            node_id (str): node unique identifier.
        Returns:
            (:py:class:`storj.model.Contact`): contact information
        """
        self.logger.info('contact_lookup(%s)', node_id)

        return model.Contact(**self._request(
            method='GET',
            path='/contacts/%s' % node_id))

    @handle_nonhttp_errors
    def file_pointers(self, bucket_id, file_id, skip, limit, exclude=None):
        """Get list of pointers associated with a file.

        See `API buckets: GET /buckets/{id}/files/{file_id}
        <https://storj.github.io/bridge/#!/buckets/get_buckets_id_files_file_id>`_

        Args:
            bucket_id (str): bucket unique identifier.
            file_id (str): file unique identifier.
            skip (str): pointer index to start the file slice.
            limit (str): number of pointers to resolve tokens for.
            exclude (list[str]): separated list of farmer node unique identifiers to exclude from token retrieval.

        Returns:
            (generator[:py:class:`storj.model.FilePointer`]): file pointers.
        """
        self.logger.debug('bucket_files(%s, %s, %s, %s, %s)', bucket_id, file_id, skip, limit, exclude)

        if bucket_id is None:
            raise ValueError('bucket unique identifier is None')
        if file_id is None:
            raise ValueError('file unique identifier is None')
        if skip is None:
            raise ValueError('starting pointer index is None')
        if limit is None:
            raise ValueError('number of tokens to resolve is None')

        pull_token = self.token_create(bucket_id, operation='PULL')

        return self._request(
            method='GET',
            path='/buckets/%s/files/%s/?skip=%s&limit=%s' % (bucket_id, file_id, skip, limit),
            headers={'x-token': pull_token.id})

    @handle_nonhttp_errors
    def file_download(self, bucket_id, file_id):
        self.logger.info('file_pointers(%s, %s)', bucket_id, file_id)

        pointers = self.file_pointers(
            bucket_id=bucket_id, file_id=file_id)

        file_contents = BytesIO()
        for pointer in pointers:
            ws = web_socket.Client(
                pointer=pointer, file_contents=file_contents)
            ws.connect()
            ws.run_forever()

        return file_contents

    @handle_nonhttp_errors
    def file_metadata(self, bucket_id, file_id):
        """Get file metadata.
        See `API buckets: GET /buckets/{id}/files/{file_id}/info
        <https://storj.github.io/bridge/#!/buckets/get_buckets_id_files_file_id_info>`_
        Args:
            bucket_id (str): bucket unique identifier.
            file_id (str): file unique identifier.
        Returns:
            (:py:class:`storj.model.File`): file metadata.
        """

        self.logger.info('file_metadata(%s, %s, %s)', bucket_id, file_id)

        response = self._request(
            method='GET',
            path='/buckets/%s/files/%s/info' % (bucket_id, file_id))

        if response is not None:
            return model.File(**response)

    @handle_nonhttp_errors
    def file_upload(self, bucket_id, file, frame):
        """Upload file.
        See `API buckets: POST /buckets/{id}/files
        <https://storj.github.io/bridge/#!/buckets/post_buckets_id_files>`_
        Args:
            bucket_id (str): bucket unique identifier.
            file (:py:class:`storj.model.File`): file to be uploaded.
            frame (:py:class:`storj.model.Frame`): frame used to stage file.
        """
        self.logger.info('file_upload(%s, %s, %s)', bucket_id, file, frame)

        def get_size(file_like_object):
            return os.stat(file_like_object.name).st_size

        file_size = get_size(file)

        # TODO:
        # encrypt file
        # shard file

        push_token = self.token_create(bucket_id, 'PUSH')

        self.logger.debug('file_upload() push_token=%s', push_token)

        # upload shards to frame
        # delete encrypted file

        self._request(
            method='POST', path='/buckets/%s/files' % bucket_id,
            # files={'file' : file},
            headers={
                #    'x-token': push_token.id,
                #    'x-filesize': str(file_size)}
                'frame': frame.id,
                'mimetype': file.mimetype,
                'filename': file.filename,
            })

    @handle_nonhttp_errors
    def file_remove(self, bucket_id, file_id):
        """Delete a file pointer from a specified bucket.
        See `API buckets: DELETE /buckets/{id}/files/{file_id}
        <https://storj.github.io/bridge/#!/buckets/delete_buckets_id_files_file_id>`_
        Args:
            bucket_id (str): bucket unique identifier.
            file_id (str): file unique identifier.
        """
        self.logger.info('file_remove(%s, %s)', bucket_id, file_id)

        self._request(
            method='DELETE',
            path='/buckets/%s/files/%s' % (bucket_id, file_id))

    @handle_nonhttp_errors
    def frame_add_shard(self, shard, frame_id):
        """Adds a shard item to the staging frame and negotiates a storage contract.

        See `API frames: PUT /frames/{frame_id}
        <https://storj.github.io/bridge/#!/frames/put_frames_frame_id>`_

        Args:
            shard (:py:class:`storj.models.Shard`): the shard.
            frame_id (str): the frame unique identifier.
        """
        self.logger.info('frame_add_shard(%s, %s)', shard, frame_id)

        data = {
            'hash': shard.hash,
            'size': shard.size,
            'index': shard.index,
            'challenges': shard.challenges,
            'tree': shard.tree.leaves,
        }

        response = self._request(
            method='PUT',
            path='/frames/%s' % frame_id,
            json=data)

        return response

    @handle_nonhttp_errors
    def file_mirrors(self, bucket_id, file_id):
        """Get list of established and available mirrors associated with a file.
        Args:
            bucket_id (str): bucket unique identifier.
            file_id (str): file unique identifier.
        Returns:
            (generator[:py:class:`storj.model.FileMirrors`]): list of mirrors of give file.
        """
        # print "test"
        self.logger.info('file_mirrors(%s, %s)', bucket_id, file_id)

        pull_token = self.token_create(bucket_id, operation='PULL')

        response = self._request(
            method='GET',
            path='/buckets/%s/files/%s/mirrors/' % (bucket_id, file_id),
            headers={'x-token': pull_token.id})

        if response is not None:
            for kwargs in response:
                yield model.FileMirrors(**kwargs)
        else:
            raise StopIteration

    @handle_nonhttp_errors
    def frame_create(self):
        """Creates a file staging frame.
        See `API frames: POST /frames
        <https://storj.github.io/bridge/#!/frames/post_frames>`_
        Returns:
            (:py:class:`storj.model.Frame`): the frame.
        """
        self.logger.info('frame_create()')

        response = self._request(
            method='POST',
            path='/frames')

        if response is not None:
            return model.Frame(**response)

    @handle_nonhttp_errors
    def frame_delete(self, frame_id):
        """Destroys the file staging frame by it's unique ID.
        See `API frames: DELETE	/frames/{frame_id}
        <https://storj.github.io/bridge/#!/frames/delete_frames_frame_id>`_
        Args:
            frame_id (str): unique identifier.
        """
        self.logger.info('frame_delete(%s)', frame_id)

        self._request(
            method='DELETE',
            path='/frames/%s' % frame_id,
            json={'frame_id': frame_id})

    @handle_nonhttp_errors
    def frame_get(self, frame_id):
        """Fetches the file staging frame by it's unique ID.
        See `API frame: GET /frames/{frame_id}
        <https://storj.github.io/bridge/#!/frames/get_frames_frame_id>`_
        Args:
            frame_id (str): unique identifier.
        Returns:
            (:py:class:`storj.model.Frame`): a frame.
        """
        self.logger.info('frame_get(%s)', frame_id)

        response = self._request(
            method='GET',
            path='/frames/%s' % frame_id,
            json={'frame_id': frame_id})

        if response is not None:
            return model.Frame(**response)

    @handle_nonhttp_errors
    def frame_list(self):
        """Returns all open file staging frames.
        See `API frame: GET /frames
        < https://storj.github.io/bridge/#!/frames/get_frames>`_
        Returns:
            (generator[:py:class:`storj.model.Frame`]): all open file staging frames.
        """
        self.logger.info('frame_list()')

        response = self._request(
            method='GET',
            path='/frames')

        if response is not None:
            for kwargs in response:
                yield model.Frame(**kwargs)
        else:
            raise StopIteration

    @handle_nonhttp_errors
    def key_delete(self, public_key):
        """Removes a public ECDSA keys.
        See `API keys: DELETE /keys/{pubkey}
        <https://storj.github.io/bridge/#!/keys/delete_keys_pubkey>`_
        Args:
            public_key (str): key to be removed.
        """
        self.logger.info('key_delete(%s)', public_key)
        self._request(
            method='DELETE',
            path='/keys/%s' % public_key)

    def key_dump(self):
        self.logger.info('key_dump()')

        if self.private_key is not None and \
                self.public_key is not None:
            print('Local Private Key: %s\nLocal Public Key: %s' % (self.private_key, self.public_key))

        keys = self.key_list()

        if not keys:
            print('No keys associated with this account.')
        else:
            print('Public keys for this account: '
                  + str([key['id'] for key in keys]))

    def key_export(self):
        self.logger.info('key_export()')

        print('Writing your public key to file...')
        with open('public.pem', 'wb') as keyfile:
            keyfile.write(self.public_key.to_pem())

        print('Writing private key to file... Keep this secret!')
        with open('private.pem', 'wb') as keyfile:
            keyfile.write(self.private_key.to_pem())

        print('Wrote keyfiles to dir: %s' % os.getcwd())

    def key_generate(self):

        self.logger.info('key_generate()')

        print("This will replace your public and private keys in 3 seconds...")
        time.sleep(3)

        self.private_key = SigningKey.generate(curve=SECP256k1, hashfunc=sha256)
        self.public_key = self.private_key.get_verifying_key()

        s = raw_input('Export keys to file for later use? [Y/N]')
        if 'Y' in s.upper():
            self.key_export()

        self.key_register(self.public_key)

    def key_import(self, private_keyfile_path, public_keyfile_path):
        self.logger.info(
            'key_import(%s, %s)',
            private_keyfile_path,
            public_keyfile_path)

        with open(public_keyfile_path, 'r') as f:
            self.public_key = VerifyingKey.from_pem(f.read())

        with open(private_keyfile_path, 'r') as f:
            self.private_key = SigningKey.from_pem(f.read())

        self.key_register(self.public_key)

    @handle_nonhttp_errors
    def key_list(self):
        """Lists the public ECDSA keys associated with the user.
        See `API keys: GET /keys
        <https://storj.github.io/bridge/#!/keys/get_keys>`_
        Returns:
            (list[str]): public keys.
        """
        self.logger.info('key_list()')

        return [kwargs['key'] for kwargs in self._request(
            method='GET',
            path='/keys'
        )]

    @handle_nonhttp_errors
    def key_register(self, public_key):
        """Register an ECDSA public key.
        See `API keys: POST /keys
        <https://storj.github.io/bridge/#!/keys/post_keys>`_
        Returns:
            (list[:py:class:`storj.model.Key`]): public keys.
        """
        self.logger.info('key_register(%s)', public_key)

        self._request(
            method='POST',
            path='/keys',
            json={'key': ecdsa_to_hex(str(public_key))})

    @handle_nonhttp_errors
    def token_create(self, bucket_id, operation):
        """Creates a token for the specified operation.
        See `API buckets: POST /buckets/{id}/tokens
        <https://storj.github.io/bridge/#!/buckets/post_buckets_id_tokens>`_
        Args:
            bucket_id (str): bucket unique identifier.
            operation (str): operation.
        Returns:
            (dict): ...
        """
        self.logger.info('token_create(%s, %s)', bucket_id, operation)

        return model.Token(**self._request(
            method='POST',
            path='/buckets/%s/tokens' % bucket_id,
            json={'operation': operation}))

    @handle_nonhttp_errors
    def send_exchange_report(self, exchange_report_data):
        """Send exchange report to bridge
            Args:
                exchange_report_data (ExchangeReport): exchange report datails.
            Returns:
                (dict): ...
               """
        self.logger.info('send_exchenge_report()')

        data = {
            'dataHash': exchange_report_data.dataHash,
            'reporterId': exchange_report_data.reporterId,
            'farmerId': exchange_report_data.farmerId,
            'clientId': exchange_report_data.clientId,
            'exchangeStart': exchange_report_data.exchangeStart,
            'exchangeEnd': exchange_report_data.exchangeEnd,
            'exchangeResultCode': exchange_report_data.exchangeResultCode,
            'exchangeResultMessage': exchange_report_data.exchangeResultMessage,
        }

        return model.ExchangeReport(**self._request(
            method='POST',
            path='/reports/exchanges',
            json=data))

    @handle_nonhttp_errors
    def user_activate(self, token):
        """Activate user.
        See `API users: GET /activations/{token}
        <https://storj.github.io/bridge/#!/users/get_activations_token>`_
        Args:
            token (str): activation token.
        """
        self.logger.info('user_activate(%s)', token)

        self._request(
            method='GET',
            path='/activations/%s' % token)

    def check_file_existence_in_bucket(self, bucket_id, filepath, file_id=None):
        # checking if file with same name or hash exist in bucket

        with open(filepath, mode='rb') as file:  # b is important -> binary
            fileContent = file.read()
        bname = (os.path.split(filepath))[1]
        file_metadata = self.file_metadata(bucket_id, file_id)
        file_hash = model.ShardManager.hash(fileContent)

        if file_metadata.filename == bname:
            return 1  # same file name
        elif file_metadata.hash == file_hash:
            return 2  # same file RIPEMD160 hash - same file content
        else:
            return False

    @handle_nonhttp_errors
    def user_activation_email(self, email, token):
        """Send user activation email.
        See `API users: POST /activations/{token}
        <https://storj.github.io/bridge/#!/users/post_activations_token>`_
        Args:
            email (str): user's email address.
            token (str): activation token.
       """
        self.logger.info('user_activation_email(%s, %s)', email, token)

        self._request(
            method='GET',
            path='/activations/%s' % token,
            json={
                'email': email,
            })

    @handle_nonhttp_errors
    def user_create(self, email, password):
        """Create a new user with Storj bridge.
        See `API users: POST /users
        <https://storj.github.io/bridge/#!/users/post_users>`_
        Args:
            email (str): user's email address.
            password (str): user's password.
        """
        self.logger.info('user_create(%s, %s)', email, password)

        password = sha256(password).hexdigest()

        response = self._request(
            method='POST',
            path='/users',
            json={
                'email': email,
                'password': password
            })

        return response

    @handle_nonhttp_errors
    def user_deactivate(self, token):
        """Discard activation token.
        See `API users: GET /activations/{token}
        <https://storj.github.io/bridge/#!/users/get_deactivations_token>`_
        Args:
            token (str): activation token.
        """
        self.logger.info('user_deactivate(%s)', token)

        self._request(
            method='DELETE',
            path='/activations/%s' % token)

    @handle_nonhttp_errors
    def user_delete(self, email):
        """Delete user account.
        See `API users: DELETE /users/{email}
        <https://storj.github.io/bridge/#!/users/post_users>`_
        Args:
            email (str): user's email address.
        """
        self.logger.info('user_delete(%s)', email)

        self._request(
            method='DELETE',
            path='/users/%s' % email)

    @handle_nonhttp_errors
    def user_reset_password(self, email):
        """Request a password reset.
        See `API users: PATCH /users/{email}
        <https://storj.github.io/bridge/#!/users/patch_users_email>`_
        Args:
            email (str): user's email address.
        """
        self.logger.info('user_reset_password(%s)', email)

        self._request(
            method='PATCH',
            path='/users/%s' % email)

    @handle_nonhttp_errors
    def user_reset_password_confirmation(self, token):
        """Confirm a password reset request.
        See `API users: GET /resets/{token}
        <https://storj.github.io/bridge/#!/users/get_resets_token>`_
        Args:
            token (str): password reset token.
        """
        self.logger.info('user_reset_password_confirmation(%s)', token)

        self._request(
            method='GET',
            path='/resets/%s' % token)



# -*- coding: utf-8 -*-
"""Storj command-line interface package."""

import logging

import click


from .. import get_client


__logger = logging.getLogger(__name__)


@click.group()
def key():
    pass


@key.command()
def generate():
    client = get_client()


@key.command
def export():
    pass


@key.command
def load():
    pass


@click.group()
def bucket():
    pass


@bucket.command()
@click.option('--storage', default=None, help='Storage limit', type=click.INT)
@click.option('--transfer', default=None,
              help='Transfer limit', type=click.INT)
@click.argument('name', type=click.STRING)
def create(storage, transfer, name):
    """Create bucket.

    Args:
        storage (int): storage limit (in GB).
        transfer (int): transfer limit (in GB).
        name (str): bucket name.
    """
    get_client().bucket_create(name, storage=storage, transfer=transfer)
    click.echo('Bucket %s created' % name)


@bucket.command()
@click.argument('bucket_id', type=click.STRING)
def get(bucket_id):
    """Get bucket."""
    bucket = get_client().bucket_get(bucket_id)

    for attr, value in sorted(bucket.__dict__.items()):
        click.echo('%s : %s' % (attr.rjust(18), value))


@bucket.command()
def list():
    """List buckets."""
    for bucket in get_client().bucket_list():
        click.echo(
            '[info]   ID: %s, Name: %s, Storage: %d, Transfer: %d' % (
                bucket.id, bucket.name, bucket.storage, bucket.transfer)
        )


@click.group()
def file():
    pass


@file.command()
@click.argument('bucket_id', type=click.STRING)
@click.argument('file_path', type=click.File('r'))
def upload(bucket_id, file_path):
    """Upload file to a storage bucket."""
    get_client().file_upload(bucket_id, file_path)



# -*- coding: utf-8 -*-
"""Test cases for the storj.exception module."""

from .. import AbstractTestCase

from storj.exception import \
    BridgeError, \
    ClientError, \
    FarmerError, \
    HashMismatchError, \
    SuppliedTokenNotAcceptedError


class BridgeErrorTestCase(AbstractTestCase):
    """Test case for the BridgeError."""

    @staticmethod
    def assertBridgeError(error, code, message):
        """Assert BridgeError expected conditions.

        Args:
            error (:py:class:`storj.exception.BridgeError`): result.
            code (int): expected error code.
            message (str): expected error message.
        """

        assert code == error.code
        assert message == error.message
        assert '[%d] %s' % (code, message) == str(error)

    def test(self):
        expected_code = 0
        expected_message = 'error'
        error = BridgeError(expected_code, expected_message)

        self.assertBridgeError(error, expected_code, expected_message)


class ClientErrorTestCase(AbstractTestCase):
    """Test case for the ClientError."""

    @staticmethod
    def assertClientError(error, message):
        """Assert ClientError expected conditions.

        Args:
            error (:py:class:`storj.exception.BridgeError`): result.
            message (str): expected error message.
        """

        assert message == error.message
        assert message == str(error)

    def test(self):
        expected = 'error'
        error = ClientError(expected)

        self.assertClientError(error, expected)


class FarmerErrorTestCase(AbstractTestCase):
    """Test case for the FarmerError."""

    @staticmethod
    def assertFarmerError(error, code, message):
        """Assert FarmerError expected conditions.

        Args:
            error (:py:class:`storj.exception.FarmerError`): result.
            code (int): expected error code.
            message (str): expected error message.
        """

        assert code == error.code
        assert message == error.message
        assert '[%d] %s' % (code, message) == str(error)

    def test(self):
        expected_code = 0
        expected_message = 'error'
        error = FarmerError(expected_code, expected_message)

        self.assertFarmerError(error, expected_code, expected_message)


class HashMismatchErrorErrorTestCase(FarmerErrorTestCase):
    """Test case for the HashMismatchError."""

    def test(self):
        expected = ''
        error = HashMismatchError()

        self.assertFarmerError(error, HashMismatchError.CODE, expected)


class SuppliedTokenNotAcceptedErrorTestCase(FarmerErrorTestCase):
    """Test case for the SuppliedTokenNotAcceptedError."""

    def test(self):
        expected = ''
        error = SuppliedTokenNotAcceptedError()

        self.assertFarmerError(
            error, SuppliedTokenNotAcceptedError.CODE, expected)



# -*- coding: utf-8 -*-
"""Test configuration module."""

from __future__ import absolute_import

import os.path

try:
    from ConfigParser import SafeConfigParser as ConfigParser
except ImportError:
    from configparser import ConfigParser

import logging


DEFAULT_CONFIG_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'storj.cfg')


class Cache(object):
    """
    Abstract class to cache configuration settings.
    """
    settings = None


def load_configuration(config_file=DEFAULT_CONFIG_FILE):
    """
    Loads configuration.

    :param config_file: the configuration file(s).
    :type config_file: str or list of str
    """
    if not os.path.exists(config_file) or not os.path.isfile(config_file):
        msg = '%s configuration file does not exist!' % config_file
        logging.getLogger(__name__).error(msg)
        raise ValueError(msg)

    parser = ConfigParser()
    try:
        parser.read(config_file)
        Cache.settings = {}
        for section in parser.sections():
            Cache.settings[section] = dict(parser.items(section))
        logging.getLogger(__name__).info(
            '%s configuration file was loaded.', config_file)
        return Cache.settings
    except Exception as e:
        Cache.settings = None
        logging.getLogger(__name__).error(
            'Failed to load configuration from %s!', config_file)
        logging.getLogger(__name__).debug(str(e), exc_info=True)
        raise e


def get():
    """
    Returns the configuration.

    :return: the configuration.
    :rtype: object (configParser.ConfigParser)
    """
    if Cache.settings is None:
        return load_configuration()
    return Cache.settings


def reset():
    """
    Reset the configuration.
    """
    Cache.settings = None



# -*- coding: utf-8 -*-
"""Storj API module."""

from binascii import b2a_hex


def ecdsa_to_hex(ecdsa_key):
    """
    Return hexadecimal string representation of the ECDSA key.

    Args:
        ecdsa_key (bytes): ECDSA key.

    Raises:
        TypeError: if the ECDSA key is not an array of bytes.

    Returns:
        str: hexadecimal string representation of the ECDSA key.
    """
    return '04%s' % b2a_hex(ecdsa_key).decode('ascii')



# -*- coding: utf-8 -*-
"""Storj exception module."""


class BridgeError(RuntimeError):
    """Storj bridge runtime exception.

    Attributes:
        code (int): error code.
        message (str): error message.
    """

    def __init__(self, code='', message=''):
        super(RuntimeError, self).__init__()

        self.code = code
        self.message = message

    def __str__(self):
        """Returns a string representation of this error.

        Returns:
            str: string representation of this error.
        """
        return '[%s] %s' % (self.code, self.message)


class ClientError(RuntimeError):
    """Storj client runtime exception.

    Attributes:
        message (str): error message.
    """

    def __init__(self, message=''):
        super(RuntimeError, self).__init__()

        self.message = message

    def __str__(self):
        """Returns a string representation of this error.

        Returns:
            str: string representation of this error.
        """
        return self.message


class FarmerError(RuntimeError):
    """Storj farmer runtime exception.

    Attributes:
        code (int): error code.
        message (str): error message.
    """

    SUPPLIED_TOKEN_NOT_ACCEPTED = 10002
    CALCULATED_HASH_NOT_MATCH_EXPECTED_RESULT = 10003

    def __init__(self, code, message=''):
        super(RuntimeError, self).__init__()

        self.code = code
        self.message = message

    def __str__(self):
        """Returns a string representation of this error.

        Returns:
            str: string representation of this error.
        """
        return '[%s] %s' % (self.code, self.message)


class SuppliedTokenNotAcceptedError(FarmerError):
    """"""

    CODE = 10002

    def __init__(self):
        super(SuppliedTokenNotAcceptedError, self).__init__(
            code=SuppliedTokenNotAcceptedError.CODE)


class HashMismatchError(FarmerError):
    """"""

    CODE = 10003

    def __init__(self):
        super(HashMismatchError, self).__init__(
            code=HashMismatchError.CODE)


StorjBridgeApiError = BridgeError
StorjFarmerError = FarmerError



import os

import click


from six.moves.configparser import RawConfigParser


APP_NAME = 'storj'
"""(str): the application name."""

CFG_EMAIL = 'email'
"""(str): configuration parameter that holds the Storj account email address."""
CFG_PASSWORD = 'password'
"""(str): configuration parameter that holds the Storj account password."""


def read_config():
    """Reads configuration storj client configuration.

    Mac OS X (POSIX):
        ~/.foo-bar
    Unix (POSIX):
        ~/.foo-bar
    Win XP (not roaming):
        ``C:\Documents and Settings\<user>\Application Data\storj``
    Win 7 (not roaming):
        ``C:\\Users\<user>\AppData\Local\storj``

    Returns:
        (tuple[str, str]): storj account credentials (email, password).
    """

    # OSX: /Users/<username>/.storj
    cfg = os.path.join(
        click.get_app_dir(
            APP_NAME,
            force_posix=True),
        'storj.ini')

    parser = RawConfigParser()
    parser.read([cfg])

    return parser.get(APP_NAME, CFG_EMAIL), parser.get(APP_NAME, CFG_PASSWORD)



# -*- coding: utf-8 -*-
"""Test cases for the storj.api module."""

from .. import AbstractTestCase

from storj import api


class FunctionsTestCase(AbstractTestCase):
    """Test case for the module functions."""

    def test_ecdsa_to_hex(self):
        """Test ecdsa_to_hex()."""

        key = b'\xf8\xd2\xaf\xa1\xdb\xa0\xee\xfd\xf9c\x01\xcf\x0c'

        self.assertEqual('04f8d2afa1dba0eefdf96301cf0c', api.ecdsa_to_hex(key))



# -*- coding: utf-8 -*-
"""Test cases for the storj.http module."""

import mock

import storj
import storj.configuration

from .. import AbstractTestCase


class FunctionsTestCase(AbstractTestCase):
    """Test case for the module functions."""

    @mock.patch('storj.configuration.click.get_app_dir')
    @mock.patch.object(storj.configuration.RawConfigParser, 'read')
    @mock.patch.object(storj.configuration.RawConfigParser, 'get')
    def test_read_config(self, mock_get, mock_read, mock_app_dir):
        # mock instance for ConfigParser.RawConfigParser
        mock_read.return_value = None
        mock_get.return_value = ['storj']
        mock_get.side_effect = [
            'someone@example.com',
            'secret'
        ]

        mock_app_dir.return_value = '/nowhere'

        assert storj.configuration.read_config() == (
            'someone@example.com',
            'secret'
        )

        mock_read.assert_called_once_with(['/nowhere/storj.ini'])
        calls = [
            mock.call('storj', 'email'),
            mock.call('storj', 'password'),
        ]
        mock_get.assert_has_calls(calls)



# -*- coding: utf-8 -*-
"""Storj API bucket endpoint integration tests."""

import logging

import pytest


from storj.exception import BridgeError

from . import Integration


class Bucket(Integration):
    """Test case for the bucket endpoint.

    Attributes:
        bucket (:py:class:`storj.models.Bucket`: test bucket.
    """

    logger = logging.getLogger('%s.Bucket' % __name__)

    def setUp(self):
        """Setup test bucket."""
        super(Bucket, self).setUp()
        self.bucket = self.client.bucket_create(
            'integration-%s' % self.test_id)

    def tearDown(self):
        """Destroy test bucket."""
        super(Bucket, self).tearDown()
        try:
            self.client.bucket_delete(self.bucket.id)
        except BridgeError as e:
            self.logger.error(e)

    def test(self):
        """Test:

        1. after setup
        1.1 get retrieves the bucket
        1.2 list has bucket

        2. delete bucket
        2.1 get does not retrieve the bucket
        2.2 list does not have bucket
        """
        self.logger.debug('---------- test() ----------')

        self.logger.debug('1.1')
        assert self.bucket == self.client.bucket_get(self.bucket.id)

        self.logger.debug('1.2')
        assert self.bucket in self.client.bucket_list()

        self.logger.debug('2.')
        self.client.bucket_delete(self.bucket.id)

        self.logger.debug('2.1')
        with pytest.raises(BridgeError):
            self.client.bucket_get(self.bucket.id)

        self.logger.debug('2.2')
        with pytest.raises(StopIteration):
            next(self.client.bucket_list())



# -*- coding: utf-8 -*-

import os

import base64
import hashlib
import hmac
import logging
import json
import requests
import threading
import time
import model

from Crypto.Cipher import AES
from Crypto import Random

from exception import BridgeError, FarmerError, SuppliedTokenNotAcceptedError
from http import Client


class Uploader:
    """

    Attributes:
        client (:py:class:`storj.http.Client`): the Storj HTTP client.
        shared_already_uploaded (int): number of shards already uploaded.
        max_retries_contract_negotiation (int): maximum number of contract negotiation retries (default=10).
        max_retries_upload_same_farmer (int): maximum number of uploads retries to the same farmer (default=3).
    """

    __logger = logging.getLogger('%s.Uploader' % __name__)

    def __init__(
            self, email, password,
            max_retries_contract_negotiation=10,
            max_retries_upload_same_farmer=3):

        self.client = Client(email, password)
        self.shards_already_uploaded = 0
        self.max_retries_contract_negotiation = max_retries_contract_negotiation
        self.max_retries_upload_same_farmer = max_retries_upload_same_farmer

    def _calculate_hmac(self, base_string, key):
        """HMAC hash calculation and returning the results in dictionary collection.

        Args:
            base_string (): .
            key (): .
        """
        hmacs = dict()
        # --- MD5 ---
        hashed = hmac.new(key, base_string, hashlib.md5)
        hmac_md5 = hashed.digest().encode('base64').rstrip('\n')
        hmacs['MD5'] = hmac_md5
        # --- SHA-1 ---
        hashed = hmac.new(key, base_string, hashlib.sha1)
        hmac_sha1 = hashed.digest().encode('base64').rstrip('\n')
        hmacs['SHA-1'] = hmac_sha1
        # --- SHA-224 ---
        hashed = hmac.new(key, base_string, hashlib.sha224)
        hmac_sha224 = hashed.digest().encode('base64').rstrip('\n')
        hmacs['SHA-224'] = hmac_sha224
        # --- SHA-256 ---
        hashed = hmac.new(key, base_string, hashlib.sha256)
        hmac_sha256 = hashed.digest().encode('base64').rstrip('\n')
        hmacs['SHA-256'] = hmac_sha256
        # --- SHA-384 ---
        hashed = hmac.new(key, base_string, hashlib.sha384)
        hmac_sha384 = hashed.digest().encode('base64').rstrip('\n')
        hmacs['SHA-384'] = hmac_sha384
        # --- SHA-512 ---
        hashed = hmac.new(key, base_string, hashlib.sha512)
        hmac_sha512 = hashed.digest().encode('base64').rstrip('\n')
        hmacs['SHA-512'] = hmac_sha512
        return hmacs

    def _prepare_bucket_entry_hmac(self, shard_array):
        """

        Args:
            shard_array (): .
        """
        storj_keyring = model.Keyring()
        encryption_key = storj_keyring.get_encryption_key('test')
        current_hmac = ''

        for shard in shard_array:
            base64_decoded = str(base64.decodestring(shard.hash)) + str(current_hmac)
            current_hmac = self._calculate_hmac(base64_decoded, encryption_key)

        self.__logger.debug('current_hmac=%s', current_hmac)

        return current_hmac

    def createNewUploadThread(self, bucket_id, file_path, tmp_file_path):
        """

        Args:
            bucket_id:
            file_path:
            tmp_file_path:
        """

        self.bid = bucket_id
        self.file_path = file_path
        self.tmp_path = tmp_file_path

        upload_thread = threading.Thread(
            target=self.file_upload,
            args=())
        upload_thread.start()

    def createNewShardUploadThread(self, shard, chapters, frame, file_name):
        """

        Args:
            shard ():
            chapters ():
            frame ():
            file_name ():
        """

        # another worker thread for single shard uploading and
        # it will retry if download fail
        upload_thread = threading.Thread(
            target=self.upload_shard(
                shard=shard,
                chapters=chapters,
                frame=frame,
                file_name_ready_to_shard_upload=file_name),
            args=())
        upload_thread.start()

    def upload_shard(self, shard, chapters, frame, file_name_ready_to_shard_upload):
        """

        Args:
            shard:
            chapters:
            frame:
            file_name_ready_to_shard_upload:
        """

        contract_negotiation_tries = 0
        exchange_report = model.ExchangeReport()

        while self.max_retries_contract_negotiation > contract_negotiation_tries:
            contract_negotiation_tries += 1
            self.__logger.debug('Negotiating contract')
            self.__logger.debug(
                'Trying to negotiate storage contract for shard at index %s...', str(chapters))

            try:
                frame_content = self.client.frame_add_shard(shard, frame.id)

                farmerNodeID = frame_content['farmer']['nodeID']

                url = 'http://%s:%d/shards/%s?token=%s' % (
                    frame_content['farmer']['address'],
                    frame_content['farmer']['port'],
                    frame_content['hash'],
                    frame_content['token'])
                self.__logger.debug('upload_shard url=%s', url)

                # begin recording exchange report
                # exchange_report = model.ExchangeReport()

                current_timestamp = int(time.time())

                exchange_report.exchangeStart = str(current_timestamp)
                exchange_report.farmerId = str(farmerNodeID)
                exchange_report.dataHash = str(shard.hash)

                farmer_tries = 0
                response = None

                while self.max_retries_upload_same_farmer > farmer_tries:
                    farmer_tries += 1

                    try:
                        self.__logger.debug(
                            'Upload shard at index %s to %s:%d attempt #%d',
                            shard.index,
                            frame_content['farmer']['address'],
                            frame_content['farmer']['port'],
                            farmer_tries)

                        mypath = os.path.join(self.tmp_path,
                                              file_name_ready_to_shard_upload +
                                              '-' + str(chapters + 1))

                        with open(mypath, 'rb') as f:
                            response = requests.post(
                                url,
                                data=self._read_in_chunks(f, shard_index=chapters),
                                timeout=1)

                        j = json.loads(str(response.content))

                        if j.get('result') == 'The supplied token is not accepted':
                            raise SuppliedTokenNotAcceptedError()

                    except FarmerError as e:
                        self.__logger.error(e)
                        continue

                    except Exception as e:
                        self.__logger.error(e)
                        self.__logger.error(
                            'Shard upload error for to %s:%d',
                            frame_content['farmer']['address'], frame_content['farmer']['port'])
                        continue

                    self.shards_already_uploaded += 1
                    self.__logger.info(
                        'Shard uploaded successfully to %s:%d',
                        frame_content['farmer']['address'],
                        frame_content['farmer']['port'])

                    self.__logger.debug(
                        '%s shards, %s sent',
                        self.all_shards_count,
                        self.shards_already_uploaded)

                    if int(self.all_shards_count) <= int(self.shards_already_uploaded):
                        self.__logger.debug('finish upload')

                    break

                self.__logger.debug('response.content=%s', response.content)

                j = json.loads(str(response.content))
                if j.get('result') == 'The supplied token is not accepted':
                    raise SuppliedTokenNotAcceptedError()

            except BridgeError as e:
                self.__logger.error(e)

                # upload failed due to Storj Bridge failure
                self.__logger.debug('Exception raised while trying to negotiate contract: ')
                continue

            except Exception as e:
                # now send Exchange Report
                # upload failed probably while sending data to farmer
                self.__logger.error(e)
                self.__logger.error(
                    'Error occured while trying to upload shard or negotiate contract. Retrying... ')
                self.__logger.error(
                    'Unhandled exception occured while trying to upload shard or negotiate contract for'
                    'shard at index %s . Retrying...', chapters)

                current_timestamp = int(time.time())
                exchange_report.exchangeEnd = str(current_timestamp)
                exchange_report.exchangeResultCode = (exchange_report.FAILURE)
                exchange_report.exchangeResultMessage = (exchange_report.STORJ_REPORT_UPLOAD_ERROR)
                # self.client.send_exchange_report(exchange_report) # send exchange report
                continue

            # uploaded with success
            current_timestamp = int(time.time())
            # prepare second half of exchange heport
            exchange_report.exchangeEnd = str(current_timestamp)
            exchange_report.exchangeResultCode = exchange_report.SUCCESS
            exchange_report.exchangeResultMessage = exchange_report.STORJ_REPORT_SHARD_UPLOADED

            self.__logger.info('Shard %s successfully added and exchange report sent.', chapters + 1)
            # self.client.send_exchange_report(exchange_report) # send exchange report
            break

    def _read_in_chunks(self, file_object, blocksize=4096, chunks=-1, shard_index=None):
        """Lazy function (generator) to read a file piece by piece.

        Default chunk size: 1k.

        Args:
            file_object (): .
            blocksize (): .
            chunks (): .
        """

        i = 0

        while chunks:
            data = file_object.read(blocksize)
            if not data:
                break
            yield data
            i += 1

            chunks -= 1

    def file_upload(self):
        """"""

        bucket_id = self.bid
        file_path = self.file_path
        tmpPath = self.tmp_path

        self.__logger.debug('Upload %s in bucket %d', file_path, bucket_id)
        self.__logger.debug('Temp folder %s', tmpPath)

        encryption_enabled = True

        bname = os.path.split(file_path)[1]  # File name

        file_mime_type = 'text/plain'

        # Encrypt file
        self.__logger.debug('Encrypting file...')

        file_crypto_tools = FileCrypto()
        # Path where to save the encrypted file in temp dir
        file_path_ready = os.path.join(tmpPath,
                                       bname + ".encrypted")
        self.__logger.debug('file_path_ready: %s', file_path_ready)

        # begin file encryption
        file_crypto_tools.encrypt_file(
            'AES',
            file_path,
            file_path_ready,
            self.client.password)

        file_name_ready_to_shard_upload = '%s.encrypted' % bname

        self.fileisdecrypted_str = ''

        file_size = os.stat(file_path).st_size

        # Get the PUSH token from Storj Bridge
        self.__logger.debug('Get PUSH Token')

        push_token = None
        try:
            push_token = self.client.token_create(bucket_id, 'PUSH')
        except BridgeError as e:
            self.__logger.error(e)
            self.__logger.debug('PUSH token create exception')

        self.__logger.debug('PUSH Token ID %s', push_token.id)

        self.__logger.debug('Frame')
        frame = None

        try:
            frame = self.client.frame_create()
        except BridgeError as e:
            self.__logger.error(e)
            self.__logger.debug('Unhandled exception while creating file staging frame')

        self.__logger.debug('frame.id = %s', frame.id)

        # Now encrypt file

        # Now generate shards
        self.__logger.debug('Sharding started...')
        shards_manager = model.ShardManager(filepath=file_path_ready,
                                            tmp_path=self.tmp_path)
        self.all_shards_count = shards_manager.index
        self.__logger.debug('Sharding ended...')

        self.__logger.debug('There are %d shards', self.all_shards_count)

        # create file hash

        chapters = 0

        for shard in shards_manager.shards:
            self.createNewShardUploadThread(shard, chapters, frame, file_name_ready_to_shard_upload)
            chapters += 1

        # finish_upload
        self.__logger.debug('Generating HMAC...')

        hash_sha512_hmac_b64 = self._prepare_bucket_entry_hmac(shards_manager.shards)
        hash_sha512_hmac = hashlib.sha224(str(hash_sha512_hmac_b64['SHA-512'])).hexdigest()

        self.__logger.debug('Now upload file')
        data = {
            'x-token': push_token.id,
            'x-filesize': str(file_size),
            'frame': frame.id,
            'mimetype': file_mime_type,
            'filename': str(bname) + str(self.fileisdecrypted_str),
            'hmac': {
                'type': "sha512",
                'value': hash_sha512_hmac
            },
        }

        self.__logger.debug('Finishing upload')
        self.__logger.debug('Adding file %s to bucket...', bname)

        success = False
        try:
            # TODO

            # This is the actual upload_file method
            response = self.client._request(
                method='POST', path='/buckets/%s/files' % bucket_id,
                # files={'file' : file},
                headers={
                    'x-token': push_token.id,
                    'x-filesize': str(file_size),
                },
                json=data,
            )
            success = True

        except BridgeError as e:
            self.__logger.error(e)
            self.__logger.debug('Unhandled bridge exception')

        if success:
            self.__logger.debug('File uploaded successfully!')

        # delete encrypted file (if encrypted and duplicated)
        if encryption_enabled and file_path_ready != "":
            self.__logger.debug('Remove file %s', file_path_ready)
            os.remove('%s*' % file_path_ready)


class FileCrypto:
    """"""

    def encrypt_file(self, algorithm, file_path, encrypted_file_save_path, password):
        """

        Args:
            algorithm:
            file_path:
            encrypted_file_save_path:
            password:
        """

        if algorithm == 'AES':
            with open(file_path, 'rb') as in_file, open(encrypted_file_save_path, 'wb') as out_file:
                self.encrypt_file_aes(in_file, out_file, password)

    def decrypt_file(self, algorithm, file_path, decrypted_file_save_path, password):
        """

        Args:
            algorithm:
            file_path:
            decrypted_file_save_path:
            password:
        """

        if algorithm == 'AES':
            with open(file_path, 'rb') as in_file, open(decrypted_file_save_path, 'wb') as out_file:
                self.decrypt_file_aes(in_file, out_file, password)

    def derive_key_and_iv(self, password, salt, key_length, iv_length):
        """

        Args:
            password:
            salt:
            key_length:
            iv_length:
        """

        d = d_i = ''

        while len(d) < key_length + iv_length:
            d_i = hashlib.md5(d_i + password + salt).digest()
            d += d_i

        return d[:key_length], d[key_length:key_length + iv_length]

    def encrypt_file_aes(self, in_file, out_file, password, key_length=32):
        """

        Args:
            in_file:
            out_file:
            password:
            key_length:
        """

        bs = AES.block_size
        salt = Random.new().read(bs - len('Salted__'))
        key, iv = self.derive_key_and_iv(password, salt, key_length, bs)
        cipher = AES.new(key, AES.MODE_CBC, iv)
        out_file.write('Salted__' + salt)
        finished = False

        while not finished:
            chunk = in_file.read(1024 * bs)
            if len(chunk) == 0 or len(chunk) % bs != 0:
                padding_length = bs - (len(chunk) % bs)
                chunk += padding_length * chr(padding_length)
                finished = True
            out_file.write(cipher.encrypt(chunk))

    def decrypt_file_aes(self, in_file, out_file, password, key_length=32):
        """

        Args:
            in_file:
            out_file:
            password:
            key_length:
        """

        bs = AES.block_size
        salt = in_file.read(bs)[len('Salted__'):]
        key, iv = self.derive_key_and_iv(password, salt, key_length, bs)
        cipher = AES.new(key, AES.MODE_CBC, iv)
        next_chunk = ''
        finished = False

        while not finished:
            chunk, next_chunk = next_chunk, cipher.decrypt(in_file.read(1024 * bs))

            if len(next_chunk) == 0:
                padding_length = ord(chunk[-1])

                if padding_length < 1 or padding_length > bs:
                    raise ValueError('bad decrypt pad (%d)' % padding_length)

                # all the pad-bytes must be the same
                if chunk[-padding_length:] != (padding_length * chr(padding_length)):
                    # this is similar to the bad decrypt:evp_enc.c from openssl program
                    raise ValueError('bad decrypt')

                chunk = chunk[:-padding_length]
                finished = True

            out_file.write(chunk)



# -*- coding: utf-8 -*-
"""Storj model module."""

import base58
import base64
import binascii
import hashlib
import logging
import math
import io
import random
import os

import six
import strict_rfc3339
import types

from Crypto.Cipher import AES
from datetime import datetime

from micropayment_core import keys

from os import urandom

from pycoin.key.Key import Key
from pycoin.serialize import b2h
from pycoin.key.BIP32Node import BIP32Node
from pycoin.serialize.bitcoin_streamer import stream_bc_string
from pycoin.ecdsa import numbertheory, generator_secp256k1
from pycoin.encoding import to_bytes_32, from_bytes_32, double_sha256

from steenzout.object import Object

from sys import platform


class Bucket(Object):
    """Storage bucket.

    A bucket is a logical grouping of files
    which the user can assign permissions and limits to.

    Attributes:
        id (str): unique identifier.
        name (str): name.
        status (str): bucket status (Active, ...).
        user (str): user email address.
        created (:py:class:`datetime.datetime`): time when the bucket was created.
        storage (int): storage limit (in GB).
        transfer (int): transfer limit (in GB).
        pubkeys ():
    """

    def __init__(
            self, id=None, name=None, status=None, user=None,
            created=None, storage=None, transfer=None, pubkeys=None,
            publicPermissions=None, encryptionKey=None):
        self.id = id
        self.name = name
        self.status = status
        self.user = user
        self.storage = storage
        self.transfer = transfer
        self.pubkeys = pubkeys
        self.publicPermissions = publicPermissions
        self.encryptionKey = encryptionKey

        # self.files = FileManager(bucket_id=self.id)
        # self.pubkeys = BucketKeyManager(
        #     bucket=self, authorized_public_keys=self.pubkeys)
        # self.tokens = TokenManager(bucket_id=self.id)

        if created is not None:
            self.created = datetime.fromtimestamp(
                strict_rfc3339.rfc3339_to_timestamp(created))
        else:
            self.created = None


class Contact(Object):
    """Contact.

    Attributes:
        address (str): hostname or IP address.
        port (str): .
        nodeID (str): node unique identifier.
        lastSeen (str): .
        protocol (str): SemVer protocol tag.
        userAgent (str):
    """

    def __init__(
            self, address=None, port=None, nodeID=None,
            lastSeen=None, protocol=None, userAgent=None, responseTime=None, timeoutRate=None, lastTimeout=None
    ):
        self.address = address
        self.port = port
        self.nodeID = nodeID
        self.lastSeen = lastSeen
        self.protocol = protocol
        self.userAgent = userAgent
        self.responseTime = responseTime
        self.timeoutRate = timeoutRate
        self.lastTimeout = lastTimeout

    @property
    def lastSeen(self):
        return self._last_seen

    @lastSeen.setter
    def lastSeen(self, value):

        if value is not None:
            self._last_seen = datetime.fromtimestamp(
                strict_rfc3339.rfc3339_to_timestamp(value))
        else:
            self._last_seen = None


class File(Object):
    """
    Attributes:
        bucket (): bucket unique identifier.
        hash ():
        mimetype ():
        filename ():
        frame (:py:class:`storj.model.Frame`): file frame.
        size ():
        shard_manager ():
    """

    def __init__(self, bucket=None, hash=None, mimetype=None,
                 filename=None, size=None, id=None, frame=None, created=None, hmac=None, erasure=None):
        self.bucket = Bucket(id=bucket)
        self.hash = hash
        self.mimetype = mimetype
        self.filename = filename
        self.size = size
        self.erasure = erasure
        self.shard_manager = None
        self.id = id
        self.frame = Frame(id=frame)
        self.created = created
        self.hmac = hmac

    @property
    def content_type(self):
        return self.mimetype

    @property
    def name(self):
        return self.filename


class FilePointer(Object):
    """File pointer.
    Args:
        hash (str):
        token (str): token unique identifier.
        operation (str):
        channel (str):
    Attributes:
        hash (str):
        token (:py:class:`storj.model.Token`): token.
        operation (str):
        channel (str):
    """

    def __init__(self, hash=None, token=None, operation=None, channel=None):
        self.hash = hash
        self.token = Token(token=token)
        self.operation = operation
        self.channel = channel


class Frame(Object):
    """File staging frame.
    Attributes:
        id (str): unique identifier.
        created (:py:class:`datetime.datetime`):
            time when the bucket was created.
        shards (list[:py:class:`Shard`]): shards that compose this frame.
    """

    def __init__(self, id=None, created=None, shards=None, locked=None, user=None, size=None):
        self.id = id
        self.locked = locked
        self.user = user
        self.size = size

        if created is not None:
            self.created = datetime.fromtimestamp(
                strict_rfc3339.rfc3339_to_timestamp(created))
        else:
            self.created = None

        if shards is None:
            self.shards = []
        else:
            self.shards = shards


class KeyPair(object):
    """
    ECDSA key pair.
    Args:
        pkey (str): hexadecimal representation of the private key (secret exponent).
        secret (str): master password.
    Attributes:
        keypair (:py:class:`pycoin.key.Key.Key`): BIP0032-style hierarchical wallet.
    Raises:
        NotImplementedError when
            a randomness source is not found.
    """

    def __init__(self, pkey=None, secret=None):

        if secret is not None:
            pkey = format(
                BIP32Node.from_master_secret(
                    secret.encode('utf-8')
                ).secret_exponent(), "064x")

        elif pkey is None:
            try:
                pkey = format(
                    BIP32Node.from_master_secret(
                        urandom(4096)
                    ).secret_exponent(), '064x')
            except NotImplementedError as e:
                raise ValueError('No randomness source found: %s' % e)

        self.keypair = Key(secret_exponent=int(pkey, 16))

    @property
    def node_id(self):
        """(str): NodeID derived from the public key (RIPEMD160 hash of public key)."""
        return b2h(self.keypair.hash160())

    @property
    def public_key(self):
        """(str): public key."""
        return b2h(self.keypair.sec(use_uncompressed=False))

    @property
    def private_key(self):
        """(str): private key."""
        return format(self.keypair.secret_exponent(), '064x')

    @property
    def address(self):
        """(): base58 encoded bitcoin address version of the nodeID."""
        return self.keypair.address(use_uncompressed=False)

    def sign(self, message, compact=True):
        """Signs the supplied message with the private key"""
        if compact:
            fd = io.BytesIO()
            stream_bc_string(fd, bytearray('Bitcoin Signed Message:\n', 'ascii'))
            stream_bc_string(fd, bytearray(message, 'utf-8'))
            mhash = from_bytes_32(double_sha256(fd.getvalue()))

            G = generator_secp256k1
            n = G.order()

            k = from_bytes_32(os.urandom(32))
            p1 = k * G
            r = p1.x()
            if r == 0:
                raise RuntimeError("amazingly unlucky random number r")
            s = (numbertheory.inverse_mod(k, n) *
                 (mhash + (self.keypair.secret_exponent() * r) % n)) % n
            if s == 0:
                raise RuntimeError("amazingly unlucky random number s")

            y_odd = p1.y() % 2
            assert y_odd in (0, 1)

            first = 27 + y_odd + (4 if not self.keypair._use_uncompressed(False) else 0)
            sig = binascii.b2a_base64(bytearray([first]) + to_bytes_32(r) + to_bytes_32(s)).strip()

            if not isinstance(sig, str):
                # python3 b2a wrongness
                sig = str(sig, 'ascii')

            return sig
        else:
            return keys.sign_sha256(self.private_key, message)


class IdecdsaCipher(Object):
    """Tools for en-/decrypt private key to/from id_ecdsa"""

    @staticmethod
    def pad(data):
        """input data is returned as a padded multiple of 16 bytes"""
        padding = 16 - len(data) % 16
        return '%s%s' % (data, padding * chr(padding))

    @staticmethod
    def unpad(data):
        """removes padding from input data and returns unpadded data"""
        return data[0:-ord(data[-1])]

    def decrypt(self, data, key, iv):
        """Decrypt data.

        The steps are:
        1. decrypt the data
        2. remove padding
        3. decode result from hexadecimal format

        Args:
            data (str/bytes): encrypted data.
            key (str):
            iv (str):

        Returns:
            (str): original data.
        """
        aes = AES.new(key, AES.MODE_CBC, iv)
        return self.unpad(aes.decrypt(data)).decode('hex')

    def encrypt(self, data, key, iv):
        """Encrypt data.

        The steps are:
        1. encode data to hexadecimal format
        2. add padding
        3. encrypt the result

        Args:
            data (str/bytes): original data.
            key (str):
            iv (str):

        Returns:
            (str): encrypted data
        """
        aes = AES.new(key, AES.MODE_CBC, iv)
        return aes.encrypt(self.pad(data.encode('hex')))

    def EVP_BytesToKey(self, password, key_len, iv_len):
        """derives a key and IV from various parameters"""
        # equivalent to OpenSSL's EVP_BytesToKey() with count 1
        # so that we make the same key and iv as nodejs version
        m = []
        len_m = 0
        i = 0
        utf8_password = password.encode('utf-8')  # in python3 this is bytes
        while len_m < (key_len + iv_len):
            md5 = hashlib.md5()
            data = utf8_password
            if i > 0:
                data = m[i - 1] + utf8_password
            md5.update(data)
            md5_digest = md5.digest()
            m.append(md5_digest)
            len_m += len(md5_digest)
            i += 1

        ms = ''.encode('utf-8')
        for mi in m:
            ms += mi

        key = ms[:key_len]
        iv = ms[key_len:key_len + iv_len]

        return key, iv

    def simpleEncrypt(self, passphrase, data):
        """Encrypt data.

        Args:
            passphrase (str): passphrase to use for encryption.
            data (str/bytes): original data.

        Returns:
            (str): base58-encoded encrypted data.
        """
        key, iv = self.EVP_BytesToKey(passphrase, 32, 16)
        return base58.b58encode(self.encrypt(data, key, iv))

    def simpleDecrypt(self, passphrase, base58_data):
        """Decrypt data.

        Args:
            passphrase (str): passphrase to use for decryption.
            base58_data (str/bytes): base58-encoded encrypted data.

        Returns:
            (str): original data.
        """
        key, iv = self.EVP_BytesToKey(passphrase, 32, 16)
        return self.decrypt(base58.b58decode(base58_data), key, iv)


class Keyring(Object):
    def __init__(self):
        self.password = None
        self.salt = None

    def generate(self):
        user_pass = raw_input("Enter your keyring password: ")
        password = hex(random.getrandbits(512 * 8))[2:-1]
        salt = hex(random.getrandbits(32 * 8))[2:-1]

        pbkdf2 = hashlib.pbkdf2_hmac('sha512', password, salt, 25000, 512)

        key = hashlib.new('sha256', pbkdf2).hexdigest()
        IV = salt[:16]
        self.export_keyring(password, salt, user_pass)
        self.password = password
        self.salt = salt

    def get_encryption_key(self, user_pass):
        # user_pass = raw_input("Enter your keyring password: ")
        password = hex(random.getrandbits(512 * 8))[2:-1]
        salt = hex(random.getrandbits(32 * 8))[2:-1]

        pbkdf2 = hashlib.pbkdf2_hmac('sha512', password, salt, 25000, 512)

        key = hashlib.new('sha256', pbkdf2).hexdigest()
        IV = salt[:16]
        # self.export_keyring(password, salt, user_pass)
        self.password = password
        self.salt = salt

        return key

    def export_keyring(self, password, salt, user_pass):
        plain = self.pad('{"pass" : "%s", \n"salt" : "%s"\n}' % (password, salt))
        IV = hex(random.getrandbits(8 * 8))[2:-1]

        aes = AES.new(self.pad(user_pass), AES.MODE_CBC, IV)

        with open('key.b64', 'wb') as f:
            f.write(base64.b64encode(IV + aes.encrypt(plain)))

    def import_keyring(self, filepath):
        with open(filepath, 'rb') as f:
            keyb64 = f.read()

        user_pass = raw_input('Enter your keyring password: ')

        key_enc = base64.b64decode(keyb64)
        IV = key_enc[:16]
        key = AES.new(self.pad(user_pass), AES.MODE_CBC, IV)

        # returns the salt and password as a dict
        creds = eval(key.decrypt(key_enc[16:])[:-4])
        self.password = creds['pass']
        self.salt = creds['salt']
        return creds


class MerkleTree(Object):
    """Simple merkle hash tree.
    Nodes are stored as strings in rows.
    Row 0 is the root node, row 1 is its children, row 2 is their children, etc

    Args:
        leaves (list[str]/types.generator[str]):
            leaves of the tree, as hex digests

    Attributes:
        leaves (list[str]): leaves of the tree, as hex digests
        depth (int): the number of levels in the tree
        count (int): the number of nodes in the tree
        rows (list[list[str]]): the levels of the tree
    """

    def __init__(self, leaves, prehashed=True):

        self.prehashed = prehashed
        self.leaves = leaves
        self.count = 0
        self._rows = []

        self._generate()

    @property
    def depth(self):
        """Calculates the depth of the tree.

        Returns:
            (int): tree depth.
        """
        pow = 0

        while (2 ** pow) < len(self._leaves):
            pow += 1

        return pow

    @property
    def leaves(self):
        """(list[str]/types.generator[str]): leaves of the tree."""
        return self._leaves

    @leaves.setter
    def leaves(self, value):

        if value is None:
            raise ValueError('Leaves should be a list.')
        elif not isinstance(value, list) and \
                not isinstance(value, types.GeneratorType):
            raise ValueError('Leaves should be a list or a generator (%s).' % type(value))

        if self.prehashed:
            # it will create a copy of list or
            # it will create a new list based on the generator
            self._leaves = list(value)
        else:
            self._leaves = [ShardManager.hash(leaf) for leaf in value]

        if not len(self._leaves) > 0:
            raise ValueError('Leaves must contain at least one entry.')

        for leaf in self._leaves:
            if not isinstance(leaf, six.string_types):
                raise ValueError('Leaves should only contain strings.')

    def _generate(self):
        """Generate the merkle tree from the leaves"""
        self._rows = [[] for _ in range(self.depth + 1)]

        # The number of leaves should be filled with hash of empty strings
        # until the number of leaves is a power of 2.
        # See https://storj.github.io/core/tutorial-protocol-spec.html
        while len(self._leaves) < (2 ** self.depth):
            self._leaves.append(ShardManager.hash(''))

        leaf_row = self.depth
        next_branches = self.depth - 1

        self._rows[leaf_row] = self._leaves
        if not self.prehashed:
            self.count += len(self._leaves)

        # Generate each row, starting from the bottom

        while next_branches >= 0:
            self._rows[next_branches] = self._make_row(next_branches)
            self.count += len(self._rows[next_branches])
            next_branches -= 1

    def _make_row(self, depth):
        """Generate the row at the specified depth"""
        row = []

        prior = self._rows[depth + 1]

        for i in range(0, len(prior), 2):
            entry = ShardManager.hash('%s%s' % (prior[i], prior[i + 1]))
            row.append(entry)

        return row

    def get_root(self):
        """Return the root of the tree"""
        return self._rows[0][0]

    def get_level(self, depth):
        """Returns the tree row at the specified depth"""
        return self._rows[depth]


class Mirror(Object):
    """Mirror or file replica settings.

    Attributes:
        hash (str):
        mirrors (int): number of file replicas.
        status (str): current file replica status.
    """

    def __init__(self, hash=None, mirrors=None, status=None):
        self.hash = hash
        self.mirrors = mirrors
        self.status = status


class FileMirrors(Object):
    """File mirrors

    Attributes:
        available (str): list of available mirrors
        established (str): list of established
    """

    def __init__(self, available=None, established=None):
        self.established = established
        self.available = available


class Shard(Object):
    """Shard.

    Attributes:
        id (str): unique identifier.
        hash (str): hash of the data.
        size (long): size of the shard in bytes.
        index (int): numberic index of the shard in the frame.
        challenges (list[str]): list of challenge numbers
        tree (list[str]): audit merkle tree
        exclude (list[str]): list of farmer nodeIDs to exclude
    """

    def __init__(self, id=None, hash=None, size=None, index=None,
                 challenges=None, tree=None, exclude=None):
        self.id = id
        # self.path = None
        self.hash = hash
        self.size = size
        self.index = index
        self.challenges = challenges
        self.tree = tree
        self.exclude = exclude

        if challenges is not None:
            self.challenges = challenges
        else:
            self.challenges = []

        if tree is not None:
            self.tree = tree
        else:
            self.tree = []

        if exclude is not None:
            self.exclude = exclude
        else:
            self.exclude = []

    def all(self):
        return_string = 'Shard{index=%s, hash=%s, ' % (
            self.index,
            self.hash)
        return_string += 'size=%s, tree={%s}, challenges={%s}' % (
            self.size,
            ', '.join(self.tree),
            ', '.join(self.challenges))
        return return_string

    def add_challenge(self, challenge):
        """Append challenge.

        Args:
            challenge (str):.
        """
        self.challenges.append(challenge)

    def add_tree(self, tree):
        """Append tree."""
        self.tree.append(tree)

    def get_public_record(self):
        pass

    def get_private_record(self):
        pass


class ShardingException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)


class ShardManager(Object):

    __logger = logging.getLogger('%s.ShardManager' % __name__)

    global SHARD_MULTIPLES_BACK, MAX_SHARD_SIZE

    MAX_SHARD_SIZE = 4294967296  # 4Gb
    SHARD_MULTIPLES_BACK = 4

    def __init__(
            self,
            filepath,
            shard_size=None,
            tmp_path=None,
            nchallenges=2,
    ):
        self.tmp_path = tmp_path
        self.nchallenges = nchallenges
        self.shard_size = shard_size
        self.filepath = filepath

    @property
    def filepath(self):
        """(str): path to the file."""
        return self._filepath

    @filepath.setter
    def filepath(self, value):
        if not isinstance(value, six.string_types):
            raise ValueError('%s must be a string' % value)
        elif not os.path.exists(value):
            raise ValueError('%s must exist' % value)
        elif not os.path.isfile(value):
            raise ValueError('%s must be a file' % value)

        self._filepath = value
        self.index = 0
        self._make_shards()

    def get_optimal_shard_parametrs(self, file_size):
        shard_parameters = {}
        accumulator = 0
        shard_size = None

        while shard_size is None:
            shard_size = self.determine_shard_size(file_size, accumulator)
            accumulator += 1

        self.__logger.debug('shard_size=%s file_size=%s', shard_size, file_size)

        if shard_size == 0:
            shard_size = file_size

        self.__logger.debug('shard_size=%s file_size=%s', shard_size, file_size)

        shard_parameters['shard_size'] = str(shard_size)
        shard_parameters['shard_count'] = math.ceil(file_size / shard_size)
        shard_parameters['file_size'] = file_size

        return shard_parameters

    def determine_shard_size(self, file_size, accumulator):

        # Based on <https://github.com/aleitner/shard-size-calculator/blob/master/src/shard_size.c>

        hops = 0

        if file_size <= 0:
            return 0

            # if accumulator != True:
            # accumulator  = 0

        self.__logger.debug('acumulator=%s', accumulator)

        # Determine hops back by accumulator

        if accumulator - SHARD_MULTIPLES_BACK < 0:
            hops = 0
        else:
            hops = accumulator - SHARD_MULTIPLES_BACK

        # accumulator = 10
        # self.shard_size(1)

        byte_multiple = self.shard_size_const(accumulator)

        check = file_size / byte_multiple

        if check > 0 and check <= 1:
            while hops > 0 and self.shard_size_const(hops) \
                    > MAX_SHARD_SIZE:
                if hops - 1 <= 0:
                    hops = 0
                else:
                    hops = hops - 1
            return self.shard_size_const(hops)

        # Maximum of 2 ^ 41 * 8 * 1024 * 1024

        if accumulator > 41:
            return 0

    def shard_size_const(self, hops):
        return 8 * (1024 * 1024) * pow(2, hops)

    def _make_shards(self):
        """Populates the shard manager with shards."""

        self.shards = []
        self.__postfix = ''
        index = 0

        # Get the file size

        fsize = os.path.getsize(self.filepath)

        optimal_shard_parametrs = \
            self.get_optimal_shard_parametrs(fsize)

        self.__numchunks = int(optimal_shard_parametrs['shard_count'])
        self.__logger.debug('Number of chunks %d', self.__numchunks)

        try:
            f = open(self.filepath, 'rb')
        except (OSError, IOError) as e:
            raise ShardingException(str(e))

        bname = os.path.split(self.filepath)[1]

        # get chunk size
        self.__chunksize = int(float(fsize) / float(self.__numchunks))

        chunksz = self.__chunksize
        total_bytes = 0
        i = 0
        for x in range(self.__numchunks):
            chunkfilename = bname + '-' + str(x + 1) + self.__postfix

            # if reading the last section,
            # calculate correct chunk size.

            if x == self.__numchunks - 1:
                chunksz = fsize - total_bytes

            self.shard_size = chunksz

            if self.tmp_path is None:
                if platform == 'linux' or platform == 'linux2':
                    # linux
                    self.tmp_path = '/tmp'
                elif platform == 'darwin':
                    # OS X
                    self.tmp_path = '/tmp'
                elif platform == 'win32':
                    # Windows
                    self.tmp_path = 'C://Windows/temp'
            self.__logger.debug('self.tmp_path=%s', self.tmp_path)

            try:
                self.__logger.debug('Writing file %s', chunkfilename)
                data = f.read(chunksz)
                total_bytes += len(data)
                inc = len(data)

                with open(os.path.join(self.tmp_path, chunkfilename),
                          'wb') as chunkf:
                    chunkf.write(data)

                challenges = self._make_challenges(self.nchallenges)

                shard = Shard(size=self.shard_size, index=index,
                              hash=ShardManager.hash(data),
                              tree=self._make_tree(challenges, data[i:i + inc]),
                              challenges=challenges)
                # hash=ShardManager.hash(data[i:i + inc]),

                self.shards.append(shard)

                index += 1
                i += 1
            except (OSError, IOError) as e:
                self.__logger.error(e)
                continue
            except EOFError as e:
                self.__logger.error(e)
                break

        self.index = len(self.shards)

    @staticmethod
    def hash(data):
        """Returns ripemd160 of sha256 of a string as a string of hex.

        Args:
            data (str): content to be digested.

        Returns:
            (str): the ripemd160 of sha256 digest.
        """

        if not isinstance(data, six.binary_type):
            data = bytes(data.encode('utf-8'))

        return binascii.hexlify(ShardManager._ripemd160(ShardManager._sha256(data))).decode('utf-8')

    @staticmethod
    def _ripemd160(b):
        """Returns the ripemd160 digest of bytes as bytes.

        Args:
            b (str): content to be ripemd160 digested.

        Returns:
            (str): the ripemd160 digest.
        """

        return hashlib.new('ripemd160', b).digest()

    @staticmethod
    def _sha256(b):
        """Returns the sha256 digest of bytes as bytes.

        Args:
            b (str): content to be sha256 digested.

        Returns:
            (str): the sha256 digest.
        """

        return hashlib.new('sha256', b).digest()

    def _make_challenges(self, challenges=12):
        """Generates the challenge strings.

        Args:
            challenges (int): number of challenges to be generated.

        Returns:
            (list[str]): list of challenges.
        """
        return [self._make_challenge_string() for _ in xrange(challenges)]

    def _make_challenge_string(self):
        return binascii.hexlify(''.join(os.urandom(32)))

    def _make_tree(self, challenges, data):
        """Creates a Storj Merkle tree.
        Args:
            challenges (list[str]): A list of random challenges.
            data (str): data to be audited.
        Returns:
            (:py:class:`MerkleTree`): audit tree.
        """
        return MerkleTree(ShardManager.hash('%s%s' % (c, data)) for c in challenges)


class Token(Object):
    """Token.

    Args:
        token (str): token unique identifier.
        bucket (str): bucket unique identifier.
        operation ():
        expires (str): expiration date, in the RFC3339 format.
        encryptionKey (str):

    Attributes:
        id (str): token unique identifier.
        bucket (:py:class:`storj.model.Bucket`): bucket.
        operation (str):
        expires (datetime.datetime): expiration date, in UTC.
        encryptionKey (str):
    """

    def __init__(
            self, token=None, bucket=None, operation=None, expires=None,
            encryptionKey=None, id=None
    ):
        self.id = token
        self.bucket = Bucket(id=bucket)
        self.operation = operation
        self.id = id

        if expires is not None:
            self.expires = datetime.fromtimestamp(
                strict_rfc3339.rfc3339_to_timestamp(expires))
        else:
            self.expires = None

        self.encryptionKey = encryptionKey


class ExchangeReport(Object):
    def __init__(
            self, dataHash=None, reporterId=None, farmerId=None, clientId=None,
            exchangeStart=None, exchangeEnd=None, exchangeResultCode=None, exchangeResultMessage=None
    ):
        self.dataHash = dataHash
        self.reporterId = reporterId
        self.farmerId = farmerId
        self.clientId = clientId
        self.exchangeStart = exchangeStart
        self.exchangeEnd = exchangeEnd
        self.exchangeResultCode = exchangeResultCode
        self.exchangeResultMessage = exchangeResultMessage

        # result codes
        self.SUCCESS = 1000
        self.FAILURE = 1100
        self.STORJ_REPORT_UPLOAD_ERROR = "TRANSFER_FAILED"
        self.STORJ_REPORT_SHARD_UPLOADED = "SHARD_UPLOADED"
        self.STORJ_REPORT_DOWNLOAD_ERROR = "DOWNLOAD_ERROR"
        self.STORJ_REPORT_SHARD_DOWNLOADED = "SHARD_DOWNLOADED"


class StorjParametrs(Object):
    def __init__(
            self, tmpPath=None
    ):
        self.tmpPath = tmpPath



# coding: utf-8
"""Storj metadata module."""

__author__ = ','.join([
    'Daniel Hawkins',
    'Fabian Barkhau',
    'James Prestwich',
    'Marc Schubert',
    'Marco Rosa',
    'Mike Bailey',
    'Pedro Salgado',
    'Shawn Wilkinson',
    'Wiktor Jezioro'
])
__author_email__ = ';'.join([
    'hwkns@alum.mit.edu',
    'f483@storj.io',
    'james@storj.io',
    'marc.schubert@gmail.com',
    'marcor165@hotmail.it',
    'mibgranny@aol.com',
    'steenzout@ymail.com',
    'shawn+pythonsdk@storj.io',
    'lakewik@lakewik.pl'
])
__version__ = '0.3.0'



# -*- coding: utf-8 -*-
"""Storj web socket module."""

import json

from ws4py.client.threadedclient import WebSocketClient


class Client(WebSocketClient):
    """Web socket client.

    Attributes:
        json (generator[:py:class:`storj.model.FilePointer`]): file pointers.
    """

    def __init__(self, pointer, file_contents):
        assert isinstance(pointer, dict)

        # according to docs it's channel
        URI = "ws://" + pointer.get('farmer')['address']
        URI += ":" + str(pointer.get('farmer')['port'])
        self.json = pointer
        self.file_contents = file_contents
        super(Client, self).__init__(URI)

    def opened(self):
        self.send(json.dumps(self.json))

    def closed(self, code, reason=None):
        return "Closed web socket %s %s" % (code, reason)

    def received_message(self, m):
        if m.is_binary:
            self.file_contents.write(m.data)



# -*- coding: utf-8 -*-
"""Storj package."""

import io


from abc import ABCMeta

from .api import ecdsa_to_hex
from .configuration import read_config
from .http import Client
from .metadata import __version__
from .model import Bucket, File, Token


def get_client():
    """Returns a pre-configured Storj HTTP client.

    Returns:
        (:py:class:`storj.http.Client`): Storj HTTP client.
    """
    return Client(*read_config())


class BucketManager(ABCMeta):
    """Class to manage buckets."""

    @staticmethod
    def all():
        buckets_json = get_client().bucket_list()
        return [Bucket(payload) for payload in buckets_json]

    @staticmethod
    def get(bucket_id):
        bucket_json = get_client().bucket_get(bucket_id=bucket_id)
        return Bucket(bucket_json)

    @staticmethod
    def create(name, storage_limit=None, transfer_limit=None):
        """Create bucket.

        Args:
            name (str):.
            storage_limit ():.
            transfer_limit ():.
        """
        bucket_json = get_client().bucket_create(
            name=name,
            storage=storage_limit,
            transfer=transfer_limit,
        )
        return Bucket(bucket_json)

    @staticmethod
    def delete(bucket_id):
        """Remove bucket.

        Args:
            bucket_id (int): bucket unique identifier.
        """
        get_client().bucket_delete(bucket_id=bucket_id)


class BucketKeyManager:
    """

    Attributes:
        bucket ():
    """

    def __init__(self, bucket, authorized_public_keys):
        self.bucket = bucket
        self._authorized_public_keys = authorized_public_keys

    def all(self):
        """"""
        return self._authorized_public_keys

    def add(self, key):
        """"""
        if not isinstance(key, str):
            key = ecdsa_to_hex(key)

        self._authorized_public_keys.append(key)
        get_client().bucket_set_keys(
            bucket_id=self.bucket.id,
            keys=self._authorized_public_keys)

    def clear(self):
        """"""
        self._authorized_public_keys = []
        get_client().bucket_set_keys(bucket_id=self.bucket.id, keys=[])

    def remove(self, key):
        """"""
        if not isinstance(key, str):
            key = ecdsa_to_hex(key)

        self._authorized_public_keys.remove(key)
        get_client().bucket_set_keys(
            bucket_id=self.bucket.id,
            keys=self._authorized_public_keys)


class FileManager:
    """"""

    def __init__(self, bucket_id):
        self.bucket_id = bucket_id

    def _upload(self, file, frame):
        """"""
        get_client().file_upload(
            bucket_id=self.bucket_id, file=file, frame=frame)

    def all(self):
        """"""
        files_json = get_client().file_list(bucket_id=self.bucket_id)
        return [File(payload) for payload in files_json]

    def delete(self, bucket_id, file_id):
        """"""
        get_client().file_remove(self, bucket_id, file_id)

    def download(self, file_id):
        """"""
        get_client().file_download(self, bucket_id, file_hash)

    def upload(self, file, frame):
        """"""
        # Support path strings as well as file-like objects
        if isinstance(file, str):
            with io.open(file, mode='rb') as file:
                self._upload(file, frame)
        else:
            self._upload(file, frame)


class TokenManager:
    """Bucket token manager.

    Attributes:
        bucket_id (int): bucket unique identifier.
    """

    def __init__(self, bucket_id):
        self.bucket_id = bucket_id

    def create(self, operation):
        """Creates a token.

        Args:
            operation (str): operation (PUSH or PULL).
        """
        operation = operation.upper()
        assert(operation in ['PUSH', 'PULL'])
        token_json = get_client().token_create(
            bucket_id=self.bucket_id, operation=operation)
        return Token(token_json)


class UserKeyManager(ABCMeta):
    """"""

    @staticmethod
    def all():
        """"""
        keys_json = get_client().key_get()
        return [payload['key'] for payload in keys_json]

    @staticmethod
    def add(key):
        """"""
        if not isinstance(key, str):
            key = ecdsa_to_hex(key)

            get_client().key_register(key)

    @staticmethod
    def clear():
        """"""
        for key in UserKeyManager.all():
            UserKeyManager.remove(key)

    @staticmethod
    def remove(key):
        """"""
        if not isinstance(key, str):
            key = ecdsa_to_hex(key)

        get_client().key_delete(key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pip.download

from pip.req import parse_requirements


from setuptools import setup, find_packages


exec(open('storj/metadata.py').read())  # load __version__


def requirements(requirements_file):
    """Return package mentioned in the given file.
    Args:
        requirements_file (str): path to the requirements file to be parsed.
    Returns:
        (list): 3rd-party package dependencies contained in the file.
    """
    return [
        str(package.req) for package in parse_requirements(
            requirements_file, session=pip.download.PipSession())]


setup(
    name='storj',
    version=__version__,  # NOQA
    description='A Python SDK for the Storj API',
    long_description=open('README.rst').read(),
    url='http://storj.io',
    author=__author__,
    author_email=__author_email__,
    license='MIT',
    dependency_links=[],
    # package_data={'storj': ['data/*.json']},
    # include_package_data=True,
    packages=find_packages(
        exclude=('*.tests', '*.tests.*', 'tests.*', 'tests')
    ),
    install_requires=requirements('requirements.txt'),
    test_suite='tests',
    tests_require=requirements('requirements-test.txt'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP',
    ],
    keywords=','.join([
        'storj', 'bridge', 'metadisk', 'api', 'client', 'sdk', 'python'
    ]),
    extras_require={
        'cli': requirements('requirements-extra-cli.txt'),
    },
    entry_points={
        'console_scripts': [
            'storj-bucket = storj.cli:bucket',
            'storj-file = storj.cli:file',
            'storj-key = storj.cli:keys',
        ]
    }
)



# -*- coding: utf-8 -*-
"""Test cases for the storj.http module."""

import mock
import pytest
import requests


from hashlib import sha256


from storj import exception, http, model


from .. import AbstractTestCase


class ClientTestCase(AbstractTestCase):
    """Test case for the Client class."""

    def setUp(self):
        super(AbstractTestCase, self).setUp()

        self.patcher_request = mock.patch.object(http.Client, '_request')
        self.mock_request = self.patcher_request.start()

        self.email = 'email@example.com'
        self.password = 's3CR3cy'
        self.client = http.Client(self.email, self.password)
        self.password_digest = sha256(
            self.password.encode('ascii')).hexdigest()

    def tearDown(self):
        self.patcher_request.stop()

    def test_init(self):
        """Test Client.__init__()."""
        assert self.email == self.client.email
        assert self.password_digest == self.client.password

    def test_add_basic_auth(self):
        """Test Client._add_basic_auth()."""
        request_kwargs = dict(headers={})
        self.client._add_basic_auth(request_kwargs)

        assert 'Authorization' in request_kwargs['headers']
        assert request_kwargs['headers']['Authorization'].startswith(b'Basic ')
        assert request_kwargs['headers']['Authorization'].endswith(b'==')

    def test_bucket_create(self):
        """Test Client.bucket_create()."""
        test_json = {'name': 'Test Bucket', 'storage': 25, 'transfer': 39}
        self.mock_request.return_value = test_json

        bucket = self.client.bucket_create(
            test_json['name'],
            storage=test_json['storage'],
            transfer=test_json['transfer']
        )

        self.mock_request.assert_called_once_with(
            method='POST',
            path='/buckets',
            json=test_json)

        assert bucket is not None
        assert isinstance(bucket, model.Bucket)

    def test_bucket_delete(self):
        """Test Client.bucket_delete()."""
        bucket_id = '57fd385426adcf743b3d39c5'
        self.client.bucket_delete(bucket_id)

        response = self.mock_request.assert_called_once_with(
            method='DELETE',
            path='/buckets/%s' % bucket_id)

        assert response is None

    def test_bucket_files(self):
        """Test Client.bucket_files()."""
        test_bucket_id = "57fd385426adcf743b3d39c5"

        response = self.client.bucket_files(test_bucket_id)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/buckets/%s/files/' % (test_bucket_id))

        self.assertIsNotNone(response)

    def test_bucket_get(self):
        """Test Client.bucket_get()."""
        test_bucket_id = "57fd385426adcf743b3d39c5"
        test_json = {'name': 'Test Bucket', 'storage': 25, 'transfer': 39}

        self.mock_request.return_value = test_json

        bucket = self.client.bucket_get(test_bucket_id)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/buckets/%s' % test_bucket_id)
        self.assertIsInstance(bucket, model.Bucket)

    def test_bucket_get_not_found(self):
        """Test Client.bucket_get() when bucket does not exist."""

        mock_error = requests.HTTPError()
        mock_error.response = mock.Mock()
        mock_error.response.status_code = 404

        self.mock_request.side_effect = mock_error

        bucket = self.client.bucket_get('inexistent')

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/buckets/inexistent')

        assert bucket is None

    def test_bucket_get_error(self):
        """Test Client.bucket_get() when a bridge error occursr."""

        mock_error = requests.HTTPError()
        mock_error.response = mock.Mock()
        mock_error.response.status_code = 500

        self.mock_request.side_effect = mock_error

        with pytest.raises(exception.BridgeError):
            self.client.bucket_get('error')

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/buckets/error')

    def test_bucket_list(self):
        """Test Client.bucket_list()."""

        # see https://storj.github.io/bridge/#!/buckets/get_buckets
        test_response = [
            {'name': 'Test Bucket 1', 'storage': 25, 'transfer': 39},
            {'name': 'Test Bucket 2', 'storage': 19, 'transfer': 83},
            {'name': 'Test Bucket 3', 'storage': 86, 'transfer': 193}]

        self.mock_request.return_value = test_response

        buckets = self.client.bucket_list()

        assert buckets is not None
        for bucket in buckets:
            assert isinstance(bucket, model.Bucket)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/buckets')

    def test_bucket_set_keys(self):
        """Test Client.bucket_set_keys()."""

        # see https://storj.github.io/bridge/#!/buckets/patch_buckets_id
        test_bucket_id = '57fd385426adcf743b3d39c5'
        test_bucket_name = 'test'
        test_keys = ['key1', 'key2', 'key3']

        self.client.bucket_set_keys(
            test_bucket_id, test_bucket_name, test_keys)

        response = self.mock_request.assert_called_once_with(
            method='PATCH',
            path='/buckets/%s' % test_bucket_id,
            json={
                'name': test_bucket_name,
                'pubkeys': test_keys
            })

        assert response is None

    def test_bucket_set_mirrors(self):
        """Test Client.bucket_set_mirrors()."""

        # see https://storj.github.io/bridge/#!/buckets/post_buckets_id_mirrors
        test_bucket_id = '57fd385426adcf743b3d39c5'
        test_file_id = '507f1f77bcf86cd799439011'

        mock_response = {
            'hash': 'fde400fe0b6a5488e10d7317274a096aaa57914d',
            'mirrors': 3,
            'status': 'pending'
        }

        self.mock_request.return_value = mock_response

        mirror = self.client.bucket_set_mirrors(
            test_bucket_id, test_file_id, 3)

        self.mock_request.assert_called_once_with(
            method='POST',
            path='/buckets/%s/mirrors' % test_bucket_id,
            json={
                'file': test_file_id,
                'redundancy': 3})

        assert mirror is not None
        assert isinstance(mirror, model.Mirror)

    def test_contact_list(self):
        """Test Client.contact_list()."""
        test_response = [
            {'protocol': '0.9.0', 'userAgent': '4.0.2'},
            {'protocol': '0.8.0', 'userAgent': '4.0.3'}
        ]

        self.mock_request.return_value = test_response

        contacts = self.client.contact_list()

        assert contacts is not None
        for contact in contacts:
            assert isinstance(contact, model.Contact)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/contacts')

    def test_contact_lookup(self):
        """Test Client.contact_lookup()."""
        test_response = {
            'protocol': '0.8.0', 'userAgent': '4.0.3'
        }

        self.mock_request.return_value = test_response

        contact = self.client.contact_lookup('node_id')

        assert contact is not None
        assert isinstance(contact, model.Contact)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/contacts/%s' % 'node_id')

    @mock.patch.object(http.Client, 'token_create')
    def test_file_pointers(self, mock_token_create):
        """Test Client.file_pointers()."""
        test_bucket_id = '1234'
        test_file_id = '5678'
        test_limit = '10'
        test_skip = '0'

        mock_token_create.return_value = model.Token(token='test_token')

        response = self.client.file_pointers(
            test_bucket_id, test_file_id, test_skip, test_limit)

        assert response is not None
        for pointer in response:
            assert isinstance(pointer, model.FilePointer)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/buckets/%s/files/%s/?skip=%s&limit=%s' % (
                test_bucket_id, test_file_id, test_skip, test_limit),
            headers={'x-token': None})

        mock_token_create.assert_called_once_with(
            test_bucket_id, operation='PULL')

    def test_file_download(self):
        """Test Client.file_download()."""
        pass

    def test_file_metadata(self):
        """Test Client.file_metadata()."""
        test_bucket_id = 'lkh39d'
        test_file_id = '72393'

        # https://storj.github.io/bridge/#!/buckets/get_buckets_id_files_file_id_info
        self.mock_request.return_value = {
            'bucket': '507f1f77bcf86cd799439011',
            'mimetype': 'video/mpeg',
            'filename': 'big_buck_bunny.mp4',
            'frame': '507f1f77bcf86cd799439191',
            'id': '507f1f77bcf86cd799430909',
            'size': 5071076
        }

        metadata = self.client.file_metadata(test_bucket_id, test_file_id)

        assert metadata is not None
        assert isinstance(metadata, model.File)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/buckets/%s/files/%s/info' % (test_bucket_id, test_file_id)
        )

    def test_file_upload(self):
        """Test Client.file_upload()."""
        # file_upload is still TODO
        pass

    def test_file_remove(self):
        """Test Client.file_remove()."""
        test_bucket_id = 'lkh39d'
        test_file_id = '72393'

        self.client.file_remove(test_bucket_id, test_file_id)

        self.mock_request.assert_called_once_with(
            method='DELETE',
            path='/buckets/%s/files/%s' % (test_bucket_id, test_file_id)
        )

    def test_frame_add_shard(self):
        """Test Client.frame_add_shard()."""
        test_tree = mock.MagicMock()
        test_shard = model.Shard(
            hash='5775772',
            index=7,
            challenges=['0118', 999, 88199, 9119, 725, 3],
            tree=test_tree,
            size=3810)
        test_json = {
            'hash': '5775772',
            'index': 7,
            'challenges': ['0118', 999, 88199, 9119, 725, 3],
            'tree': test_tree.leaves,
            'size': 3810}

        test_frame_id = '8193'

        self.client.frame_add_shard(test_shard, test_frame_id)

        self.mock_request.assert_called_once_with(
            method='PUT',
            path='/frames/%s' % test_frame_id,
            json=test_json)

    def test_frame_create(self):
        """Test Client.frame_create()."""

        frame = self.client.frame_create()

        self.mock_request.assert_called_once_with(
            method='POST',
            path='/frames')

        assert frame is not None
        assert isinstance(frame, model.Frame)

    def test_frame_delete(self):
        """Test Client.frame_delete()."""
        test_frame_id = '314159265358979265'
        test_json = {
            'frame_id': test_frame_id}

        self.client.frame_delete(test_frame_id)

        self.mock_request.assert_called_once_with(
            method='DELETE',
            path='/frames/%s' % test_frame_id,
            json=test_json
        )

    def test_frame_get(self):
        """Test Client.frame_get()."""
        test_frame_id = '1234'
        test_json = {
            'created': '2016-03-04T17:01:02.629Z',
            'id': '507f1f77bcf86cd799439011',
            'shards': [{
                'hash': 'fde400fe0b6a5488e10d7317274a096aaa57914d',
                'size': 4096,
                'index': 0}]}

        self.mock_request.return_value = test_json

        frame = self.client.frame_get(test_frame_id)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/frames/%s' % test_frame_id,
            json={'frame_id': test_frame_id})

        assert frame is not None
        assert isinstance(frame, model.Frame)

    def test_frame_list(self):
        """Test Client.frame_list()."""

        # see https://storj.github.io/bridge/#!/frames/get_frames
        self.mock_request.return_value = [{
            'created': '2016-03-04T17:01:02.629Z',
            'id': '507f1f77bcf86cd799439011'
        }]

        response = self.client.frame_list()

        assert response is not None
        for frame in response:
            assert isinstance(frame, model.Frame)

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/frames')

    def test_key_delete(self):
        """Test Client.key_delete()."""
        test_key = '39ddkakdi'

        self.client.key_delete(test_key)

        self.mock_request.assert_called_once_with(
            method='DELETE',
            path='/keys/%s' % test_key
        )

    @mock.patch.object(http.Client, 'key_list')
    @mock.patch('sys.stdout', autospec=True)
    def test_key_dump(self, mock_stdout, mock_key_list):
        """Test Client.key_dump()."""
        self.client.private_key = '1234'
        self.client.public_key = '5678'
        test_keys = [
            {'id': 7},
            {'id': 8}
        ]

        mock_key_list.return_value = test_keys

        self.client.key_dump()

        mock_key_list.assert_called_once_with()

        calls = [
            mock.call("Local Private Key: %s" % self.client.private_key
                      + "\nLocal Public Key: %s" % self.client.public_key),
            mock.call("\n"),
            mock.call(
                "Public keys for this account: "
                + str([key['id'] for key in test_keys])),
            mock.call("\n")]
        mock_stdout.write.assert_has_calls(calls)

    @mock.patch.object(http.Client, 'key_list', return_value=[])
    @mock.patch('sys.stdout', autospec=True)
    def test_key_dump_missing_keys(self, mock_stdout, mock_key_list):
        """Test Client.key_dump() with missing keys."""
        self.client.private_key = None

        self.client.key_dump()

        mock_key_list.assert_called_once_with()
        self.client.logger.info(mock_stdout.write.call_args_list)

        calls = [
            mock.call('No keys associated with this account.'),
            mock.call('\n')]

        mock_stdout.write.assert_has_calls(calls)

    @mock.patch('storj.http.os', autospec=True)
    @mock.patch('storj.http.open', create=True)
    @mock.patch('sys.stdout', autospec=True)
    def test_key_export(self, mock_stdout, mock_open, mock_os):
        """Test Client.key_export()."""
        test_cwd = '~/.keys/'

        mock_os.getcwd.return_value = test_cwd

        mock_file = mock.MagicMock()
        mock_open.return_value = mock_file
        mock_file_handle = mock_open.return_value.__enter__.return_value

        self.client.public_key = mock.MagicMock()
        self.client.public_key.to_pem.return_value = '7'
        self.client.private_key = mock.MagicMock()
        self.client.private_key.to_pem.return_value = '8'

        self.client.key_export()

        open_calls = [
            mock.call('public.pem', 'wb'),
            mock.call('private.pem', 'wb')]
        file_write_calls = [
            mock.call('7'),
            mock.call('8')]
        print_calls = [
            mock.call('Writing your public key to file...'),
            mock.call('\n'),
            mock.call('Writing private key to file... Keep this secret!'),
            mock.call('Wrote keyfiles to dir: ' + test_cwd)]

        mock_open.assert_has_calls(open_calls, any_order=True)
        mock_file_handle.write.assert_has_calls(file_write_calls)
        mock_stdout.write.assert_has_calls(print_calls, any_order=True)

    def test_key_generate(self):
        """Test Client.key_generate()."""
        pass

    def test_key_import(self):
        """Test Client.key_import()."""
        pass

    def test_key_list(self):
        """Test Client.key_list()."""

        self.mock_request.return_value = [
            # see https://storj.github.io/bridge/#!/keys/get_keys
            {'user': 'dnr@dnr.com', 'key': 'test_key'}
        ]

        response = self.client.key_list()

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/keys'
        )

        assert response is not None
        assert response == ['test_key']

    @mock.patch('storj.http.ecdsa_to_hex')
    def test_key_register(self, mock_ecdsa):
        """Test Client.key_register()."""
        test_hex_key = 'hex encoded key'
        test_json = {'key': test_hex_key}

        mock_ecdsa.return_value = test_hex_key

        response = self.client.key_register('key')

        assert response is None

        mock_ecdsa.assert_called_once_with('key')
        self.mock_request.assert_called_once_with(
            method='POST',
            path='/keys',
            json=test_json)

    def test_token_create(self):
        """Test Client.token_create()."""

        test_bucket_id = '1234'
        test_json = {'operation': 'PULL'}

        # https://storj.github.io/bridge/#!/buckets/post_buckets_id_tokens
        self.mock_request.return_value = dict(
            token='string',
            bucket='string',
            expires='2016-10-13T04:23:48.183Z',
            operation='string',
            encryptionKey='string'
        )

        response = self.client.token_create(test_bucket_id, 'PULL')

        assert response is not None
        assert isinstance(response, model.Token)

        self.mock_request.assert_called_once_with(
            method='POST',
            path='/buckets/%s/tokens' % test_bucket_id,
            json=test_json)

    def test_user_activate(self):
        """Test Client.user_activate()."""

        self.mock_request.return_value = None

        response = self.client.user_activate('token')

        assert response is None

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/activations/token')

    def test_user_activation_email(self):
        """Test Client.user_activation_email()."""

        self.mock_request.return_value = None

        response = self.client.user_activation_email('email', 'token')

        assert response is None

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/activations/token',
            json={'email': 'email'})

    @mock.patch('storj.http.sha256')
    def test_user_create(self, mock_sha256):
        """Test Client.user_create()."""
        test_email = 'a@b.com'
        test_password = 'toast'
        test_hashed_password = 'hashed password'

        mock_sha256.return_value.hexdigest.return_value = test_hashed_password

        self.client.user_create(test_email, test_password)

        mock_sha256.assert_called_once_with(test_password)
        mock_sha256.return_value.hexdigest.assert_called_once_with()

        self.mock_request.assert_called_once_with(
            method='POST',
            path='/users',
            json={'email': test_email, 'password': test_hashed_password})

    def test_user_deactivate(self):
        """Test Client.user_deactivate()."""

        self.mock_request.return_value = None

        response = self.client.user_deactivate('token')

        assert response is None

        self.mock_request.assert_called_once_with(
            method='DELETE',
            path='/activations/token')

    def test_user_delete(self):
        """Test Client.user_delete()."""

        self.mock_request.return_value = None

        response = self.client.user_delete('email')

        assert response is None

        self.mock_request.assert_called_once_with(
            method='DELETE',
            path='/users/email')

    def test_user_reset_password(self):
        """Test Client.user_reset_password()."""

        self.mock_request.return_value = None

        response = self.client.user_reset_password('email')

        assert response is None

        self.mock_request.assert_called_once_with(
            method='PATCH',
            path='/users/email')

    def test_user_reset_password_confirmation(self):
        """Test Client.user_reset_password_confirmation()."""

        self.mock_request.return_value = None

        response = self.client.user_reset_password_confirmation('token')

        assert response is None

        self.mock_request.assert_called_once_with(
            method='GET',
            path='/resets/token')



# -*- coding: utf-8 -*-
"""Test cases for the storj package."""

import mock
import storj


from hashlib import sha256

import storj.configuration
from ecdsa import keys

from .. import AbstractTestCase


class FunctionsTestCase(AbstractTestCase):
    """Test case for the package functions."""

    @mock.patch('storj.read_config')
    def test_get_client(self, mock_read_config):
        mock_read_config.return_value = ('someone@example.com', 'secret')

        client = storj.get_client()

        assert client is not None
        assert client.email == 'someone@example.com'
        assert client.password == sha256('secret'.encode('ascii')).hexdigest()

        mock_read_config.assert_called_once_with()



# -*- coding: utf-8 -*-
"""Test cases for the storj.cli package."""

import logging
import mock


from click.testing import CliRunner
from datetime import datetime

from storj import cli, model


from .. import AbstractTestCase


class BucketTestCase(AbstractTestCase):
    """Test case for bucket commands."""

    logger = logging.getLogger('%s.Client' % __name__)
    runner = CliRunner()

    def setUp(self):
        self.timestamp = datetime.utcnow().strftime('%Y%m%d_%H%M%S')

        self.bucket = model.Bucket(
            id='id-%s' % self.timestamp, name='bucket-%s' % self.timestamp,
            storage=100, transfer=100)

        self.client_class_patch = mock.patch('storj.http.Client', spec=True)
        self.mock_client_class = self.client_class_patch.start()
        self.mock_client = self.mock_client_class()

        self.get_client_patch = mock.patch('storj.cli.get_client')
        self.mock_get_client = self.get_client_patch.start()
        self.mock_get_client.return_value = self.mock_client

        self.mock_action = mock.MagicMock()

    def tearDown(self):
        self.get_client_patch.stop()
        self.client_class_patch.stop()

    def test_bucket_create(self):
        """Test create command."""
        self.mock_action.return_value = None
        self.mock_client.bucket_create = self.mock_action

        result = self.runner.invoke(cli.create, [self.bucket.name])

        assert result.exit_code == 0
        assert result.output == 'Bucket %s created\n' % self.bucket.name

        self.mock_action.assert_called_once_with(
            self.bucket.name, storage=None, transfer=None)

    def test_bucket_create_with_options(self):
        """Test create command."""
        self.mock_action.return_value = None
        self.mock_client.bucket_create = self.mock_action

        args = [
            '--storage=%u' % self.bucket.storage,
            '--transfer=%u' % self.bucket.transfer,
            self.bucket.name
        ]

        self.logger.debug('test_bucket_create_with_options args=%s', args)
        result = self.runner.invoke(cli.create, args)

        assert result.exit_code == 0
        assert result.output == 'Bucket %s created\n' % self.bucket.name

        self.mock_action.assert_called_once_with(
            self.bucket.name,
            storage=self.bucket.storage,
            transfer=self.bucket.transfer)

    def test_bucket_get(self):
        """Test get command."""
        self.mock_action.return_value = self.bucket
        self.mock_client.bucket_get = self.mock_action

        result = self.runner.invoke(cli.get, [self.bucket.id])

        assert result.exit_code == 0
        assert result.output == '\n'.join([
            '           created : %s' % self.bucket.created,
            '     encryptionKey : %s' % self.bucket.encryptionKey,
            '                id : %s' % self.bucket.id,
            '              name : %s' % self.bucket.name,
            '           pubkeys : %s' % self.bucket.pubkeys,
            ' publicPermissions : %s' % self.bucket.publicPermissions,
            '            status : %s' % self.bucket.status,
            '           storage : %u' % self.bucket.storage,
            '          transfer : %u' % self.bucket.transfer,
            '              user : %s\n' % self.bucket.user,
        ])

        self.mock_action.assert_called_once_with(self.bucket.id)

    def test_bucket_list(self):
        """Test list command."""
        self.mock_action.return_value = [self.bucket]
        self.mock_client.bucket_list = self.mock_action

        result = self.runner.invoke(cli.list, [])

        assert result.exit_code == 0
        msg = '[info]   ID: %s, Name: %s, Storage: %d, Transfer: %d\n'
        assert result.output == msg % (
            self.bucket.id, self.bucket.name,
            self.bucket.storage, self.bucket.transfer
        )

        self.mock_action.assert_called_once_with()



# -*- coding: utf-8 -*-
"""Test logging module."""

import os

import logging
import logging.config as config


DEFAULT_CONFIG_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'logging.conf')


def load_configuration(config_file=DEFAULT_CONFIG_FILE):
    """
    Loads logging configuration from the given configuration file.

    :param config_file:
        the configuration file (default=/etc/package/logging.conf)
    :type config_file: str
    """
    if not os.path.exists(config_file) or not os.path.isfile(config_file):
        msg = '%s configuration file does not exist!', config_file
        logging.getLogger(__name__).error(msg)
        raise ValueError(msg)

    try:
        config.fileConfig(config_file, disable_existing_loggers=False)
        logging.getLogger(__name__).info(
            '%s configuration file was loaded.', config_file)
    except Exception as e:
        logging.getLogger(__name__).error(
            'Failed to load configuration from %s!', config_file)
        logging.getLogger(__name__).debug(str(e), exc_info=True)
        raise e



# -*- coding: utf-8 -*-
"""Test cases for the storj.model module."""

import mock
import os
import pytest
import shutil
import six
import strict_rfc3339
import tempfile

from datetime import datetime

from pycoin.key.Key import Key
from pycoin.serialize import b2h
from pycoin.key.BIP32Node import BIP32Node

from storj.model import \
    Bucket, Contact, File, FilePointer, Frame, KeyPair, \
    MerkleTree, Mirror, Shard, ShardManager, Token, IdecdsaCipher

from .. import AbstractTestCase


class BucketTestCase(AbstractTestCase):
    """Test case for the Bucket class."""

    def test_init(self):
        """Test Bucket.__init__()."""
        kwargs = dict(
            created='2016-10-13T04:23:48.183Z',
            id='510b23e9f63a77d939a72a77',
            name='integration-20161013_042347',
            pubkeys=[],
            status='Active',
            storage=0,
            transfer=0,
            user='steenzout@ymail.com')

        bucket = Bucket(**kwargs)

        assert bucket.created == datetime.fromtimestamp(
            strict_rfc3339.rfc3339_to_timestamp(
                '2016-10-13T04:23:48.183Z'))
        assert bucket.id == '510b23e9f63a77d939a72a77'
        assert bucket.name == 'integration-20161013_042347'
        assert bucket.pubkeys == []
        assert bucket.status == 'Active'
        assert bucket.storage == 0
        assert bucket.transfer == 0
        assert bucket.user == 'steenzout@ymail.com'


class ContactTestCase(AbstractTestCase):
    """Test case for the Contact class."""

    def test_init(self):
        """Test Contact.__init__()."""

        kwargs = dict(
            address='api.storj.io',
            port=8443,
            nodeID='32033d2dc11b877df4b1caefbffba06495ae6b18',
            lastSeen='2016-05-24T15:16:01.139Z',
            protocol='0.7.0',
            userAgent='4.0.3'
        )

        contact = Contact(**kwargs)

        assert contact.address == kwargs['address']
        assert contact.port == kwargs['port']
        assert contact.nodeID == kwargs['nodeID']
        assert contact.lastSeen == datetime.fromtimestamp(
            strict_rfc3339.rfc3339_to_timestamp(kwargs['lastSeen']))
        assert contact.protocol == kwargs['protocol']
        assert contact.userAgent == kwargs['userAgent']


class FileTestCase(AbstractTestCase):
    """Test case for the File class."""

    def test_init(self):
        """Test File.__init__()."""

        kwargs = dict(
            bucket='bucket_id',
            hash='hash',
            mimetype='mimetype',
            filename='filename',
            frame='frame_id'
        )

        f = File(**kwargs)

        assert f.bucket == Bucket(id=kwargs['bucket'])
        assert f.hash == kwargs['hash']
        assert f.mimetype == kwargs['mimetype']
        assert f.filename == kwargs['filename']
        assert f.frame == Frame(id=kwargs['frame'])
        assert f.shard_manager is None


class FilePointerTestCase(AbstractTestCase):
    """Test case for the FilePointer class."""

    def test_init(self):
        """Test File.__init__()."""

        # https://storj.github.io/bridge/#!/buckets/get_buckets_id_files_file_id
        kwargs = dict(
            hash='ba084d3f143f2896809d3f1d7dffed472b39d8de',
            token='99cf1af00b552113a856f8ef44f58d22269389e8'
                  '009d292bafd10af7cc30dcfa',
            operation='PULL',
            channel='ws://farmer.hostname:4000'
        )

        fp = FilePointer(**kwargs)

        assert fp.hash == kwargs['hash']
        assert fp.token == Token(token=kwargs['token'])
        assert fp.operation == kwargs['operation']
        assert fp.channel == kwargs['channel']


class FrameTestCase(AbstractTestCase):
    """Test case for the Frame class."""

    def test_init(self):
        """Test Frame.__init__()."""
        kwargs = dict(
            created='2016-10-13T04:23:48.183Z',
            id='510b23e9f63a77d939a72a77',
            shards=[])

        frame = Frame(**kwargs)

        assert frame.created == datetime.fromtimestamp(
            strict_rfc3339.rfc3339_to_timestamp(
                '2016-10-13T04:23:48.183Z'))
        assert frame.id == '510b23e9f63a77d939a72a77'
        assert frame.shards == []


class KeyPairTestCase(AbstractTestCase):
    """Test case for the KeyPair class."""

    def _assert_wallet(self, wallet):

        assert isinstance(wallet.keypair, Key)
        assert wallet.private_key == self.private_key
        assert wallet.public_key == self.public_key
        assert wallet.node_id == 'abb5b062c7797924bb5b7d65e4f8a2f7c6e0311f'
        assert wallet.address == '1GevCn3H1p76GHHPS9nzg8wBYLFaS2MAmX'

    def setUp(self):
        self.master_password = 'master_password'
        self.secret = (
            'c7d360e0d7d6820ea8d33cc7ad81bf9d'
            '04c2f9c793f21cbf0a4a004350346ab8'
        )
        self.secret_exponent = int(self.secret, 16)
        self.private_key = (
            'c7d360e0d7d6820ea8d33cc7ad81bf9d'
            '04c2f9c793f21cbf0a4a004350346ab8'
        )
        self.public_key = (
            '03980352d67d91a8cf64251b1d4f72726'
            '54c70aa21932aaea1b359c47b26aee9d0'
        )

    def test_init(self):
        """Test KeyPair.__init__()."""

        self._assert_wallet(KeyPair(**dict(
            pkey=self.secret
        )))
        self._assert_wallet(KeyPair(**dict(secret=self.master_password)))


class IdecdsaCipherTestCase(AbstractTestCase):
    """Test case for the IdecdsaCipher class."""

    def test_encrypt_decrypt(self):
        password = 'testpassword'
        data = ('c7d360e0d7d6820ea8d33cc7ad81bf9d'
                '04c2f9c793f21cbf0a4a004350346ab8')

        cipher = IdecdsaCipher()

        assert cipher.simpleDecrypt(
            password, cipher.simpleEncrypt(password, data)) == \
            data

        bytes_data = 'testpassword'.encode('utf-8')
        assert cipher.simpleDecrypt(
            password, cipher.simpleEncrypt(password, bytes_data)) == \
            bytes_data

    def test_pad_unpad(self):
        data = '0123456789abcdef'

        assert data == IdecdsaCipher.unpad(IdecdsaCipher.pad(data))

        data = b'0123456789abcdef'

        assert data == IdecdsaCipher.unpad(IdecdsaCipher.pad(data))

    def test_bytes_to_key(self):
        password = 'secret'
        cipher = IdecdsaCipher()

        assert ('^\xbe"\x94\xec\xd0\xe0\xf0\x8e\xab',
                'v\x90\xd2\xa6\xeei&\xae\\\xc8T\xe3kk\xdf\xca6hH\xde') == \
            cipher.EVP_BytesToKey(password, 10, 20)


class MirrorTestCase(AbstractTestCase):
    """Test case for the Mirror class."""

    def test_init(self):
        """Test Mirror.__init__()."""

        kwargs = dict(
            hash='fde400fe0b6a5488e10d7317274a096aaa57914d',
            mirrors=3,
            status='pending')

        mirror = Mirror(**kwargs)

        assert mirror.hash == kwargs['hash']
        assert mirror.mirrors == kwargs['mirrors']
        assert mirror.status == kwargs['status']


class ShardTestCase(AbstractTestCase):
    """Test case for the Shard class."""

    def _assert_init(self, kwargs):
        """Run __init__ assertions.

        Args:
            kwargs (dict): keyword arguments for the Shard initializer.

        Raises:
            AssertionError: Shard attributes is not set as expected.
        """

        shard = Shard(**kwargs)

        if 'challenges' in kwargs:
            assert kwargs['challenges']
        else:
            assert shard.challenges == []
        if 'exclude' in kwargs:
            assert kwargs['exclude']
        else:
            assert shard.exclude == []

        assert shard.hash == kwargs['hash']
        assert shard.id == kwargs['id']
        assert shard.index == kwargs['index']

        assert shard.size is None

        assert kwargs['tree'] if 'tree' in kwargs else [] == shard.tree

    def test_init(self):
        """Test Shard.__init__()."""

        kwargs = dict(
            hash='',
            id='510b23e9f63a77d939a72a77',
            index='')
        self._assert_init(kwargs)

        kwargs = dict(
            challenges=['abc'],
            exclude=['abc'],
            hash='',
            id='510b23e9f63a77d939a72a77',
            index='',
            tree=['abc'])
        self._assert_init(kwargs)

    def test_add_challenge(self):
        """Test Shard.add_challenge()."""

        shard = Shard()
        shard.add_challenge('challenge1')

        assert shard.challenges == ['challenge1']

    def test_add_tree(self):
        """Test Shard.add_tree()."""

        shard = Shard()
        shard.add_tree('node')

        assert shard.tree == ['node']

    def test_get_public_record(self):
        pass

    def test_get_private_record(self):
        pass


class ShardManagerTestCase(AbstractTestCase):
    """Test case for the ShardManager class."""

    @mock.patch.object(ShardManager, '_make_challenges', return_value=[])
    @mock.patch.object(ShardManager, '_make_tree', return_value=[])
    def test_property_filepath(self, mock_tree, mock_challenges):
        """Test filepath property."""

        # filepath is not a str
        with pytest.raises(ValueError):
            ShardManager(1, 1)

        # filepath does not exist
        with pytest.raises(ValueError):
            ShardManager('/dev/nowhere', 1)

        # file path is a directory
        tmpdir = tempfile.mkdtemp()
        try:
            with pytest.raises(ValueError):
                ShardManager(tmpdir, 1)
        finally:
            shutil.rmtree(tmpdir)

        # file path is a text file
        content = '1234567890'
        self._assert_shard_manager(
            content, 'w+t', mock_tree, mock_challenges)

        content = b'1234567890'
        self._assert_shard_manager(
            content, 'w+b', mock_tree, mock_challenges)

    def _assert_shard_manager(self, content, mode, mock_tree, mock_challenges):
        tmpfile = tempfile.NamedTemporaryFile(mode, delete=False)
        try:
            tmpfile.write(content)
            tmpfile.flush()

            size = 10
            nchallenges = 2
            sm = ShardManager(
                tmpfile.name,
                shard_size=size,
                nchallenges=nchallenges)

            assert sm.filepath == tmpfile.name
            assert sm.shard_size == size
            assert len(sm.shards) > 0
            assert len(sm.shards) == sm.index

        finally:
            tmpfile.close()
            os.remove(tmpfile.name)

        mock_challenges.assert_called_once_with(nchallenges)
        if isinstance(content, six.binary_type):
            mock_tree.assert_called_once_with(
                mock_challenges.return_value, content)
        else:
            mock_tree.assert_called_once_with(
                mock_challenges.return_value, bytes(content.encode('utf-8')))

        mock_tree.reset_mock()
        mock_challenges.reset_mock()

    @mock.patch.object(ShardManager, '_sha256')
    @mock.patch.object(ShardManager, '_ripemd160')
    @mock.patch('storj.model.bytes')
    @mock.patch('storj.model.binascii')
    def test_hash(self,
                  mock_binascii, mock_bytes, mock_ripemd160, mock_sha256):
        """Test ShardManager.hash()"""

        hash_output = mock.MagicMock()
        test_data = mock.MagicMock()
        test_data.encode.return_value = test_data

        mock_ripemd160.return_value = hash_output
        mock_sha256.return_value = hash_output

        mock_bytes.return_value = test_data
        mock_binascii.hexlify.return_value = test_data

        output = ShardManager.hash(test_data)

        assert output is not None

        test_data.encode.assert_called_with('utf-8')
        mock_bytes.assert_called_with(test_data)
        mock_sha256.assert_called_with(test_data)
        mock_ripemd160.assert_called_with(hash_output)
        mock_binascii.hexlify.assert_called_with(hash_output)
        test_data.decode.assert_called_with('utf-8')

    @mock.patch('storj.model.hashlib')
    def test_ripemd160_binary(self, mock_hashlib):
        """Test ShardManager._ripemd160"""
        test_data = b'ab'

        output = ShardManager._ripemd160(test_data)

        assert output is not None

        mock_hashlib.new.assert_called_with('ripemd160', test_data)
        mock_hashlib.new.return_value.digest.assert_called_once_with()

    @mock.patch('storj.model.hashlib')
    def test_ripemd160_text(self, mock_hashlib):
        """Test ShardManager._ripemd160"""

        test_data = 'ab'

        output = ShardManager._ripemd160(test_data)

        assert output is not None

        mock_hashlib.new.assert_called_with('ripemd160', test_data)
        mock_hashlib.new.return_value.digest.assert_called_once_with()

    @mock.patch('storj.model.hashlib')
    def test_sha256_binary(self, mock_hashlib):
        """Test ShardManager._sha256"""
        test_data = b'ab'

        output = ShardManager._ripemd160(test_data)

        mock_hashlib.new.assert_called_with('ripemd160', test_data)
        mock_hashlib.new.return_value.digest.assert_called_once_with()

    @mock.patch('storj.model.hashlib')
    def test_sha256_text(self, mock_hashlib):
        """Test ShardManager._sha256"""
        test_data = 'ab'

        output = ShardManager._ripemd160(test_data)

        mock_hashlib.new.assert_called_with('ripemd160', test_data)
        mock_hashlib.new.return_value.digest.assert_called_once_with()


class TokenTestCase(AbstractTestCase):
    """Test case for the Token class."""

    def _assert_init(self, kwargs):
        """Run __init__ assertions.

        Args:
            kwargs (dict): keyword arguments for the Token initializer.

        Raises:
            AssertionError: Token attributes is not set as expected.
        """

        token = Token(**kwargs)

        if 'bucket' in kwargs:
            assert token.bucket_id == kwargs['bucket']
        else:
            assert token.bucket_id is None

        if 'expires' in kwargs:
            assert token.expires == datetime.fromtimestamp(
                strict_rfc3339.rfc3339_to_timestamp(
                    kwargs['expires']))
        else:
            assert token.expires is None

        if 'operation' in kwargs:
            assert token.operation == kwargs['operation']
        else:
            assert token.operation is None

        if 'token' in kwargs:
            assert token.id == kwargs['token']
        else:
            assert token.id is None

    def test_init(self):
        """Test Token.__init__()."""
        kwargs = dict(
            bucket='',
            expires='2016-10-13T04:23:48.183Z',
            operation='unknown',
            token='510b23e9f63a77d939a72a77')
        self._assert_init(kwargs)

        kwargs = dict(
            bucket='',
            expires='2016-10-13T04:23:48.183Z',
            operation='unknown',
            token='510b23e9f63a77d939a72a77')
        self._assert_init(kwargs)


class MerkleTreeTestCase(AbstractTestCase):
    """Test case for the MerkleTree Class"""

    def setUp(self):
        super(AbstractTestCase, self).setUp()

        self.leaves = ['a', 'b', 'c', 'd']
        self.tree = MerkleTree(self.leaves)

    def _assert_init(self, leaves, mock_generate, kwargs):
        """Run init assertions for MerkleTree."""

        tree = MerkleTree(leaves, **kwargs)

        assert tree.leaves == ['a', 'b']

        prehashed = kwargs['prehashed'] if 'prehashed' in kwargs else True
        assert prehashed == tree.prehashed

        assert tree.count == 0
        assert tree._rows == []
        assert tree.depth == 1

        mock_generate.assert_called_once_with()
        mock_generate.reset_mock()

    @mock.patch.object(MerkleTree, '_generate')
    def test_init(self, mock_generate):
        """Test MerkleTree.__init__()."""

        # success
        for leaves, kwargs in (
                (['a', 'b'], dict()),
                (['a', 'b'], dict(prehashed=True)),
                ((x for x in ['a', 'b']), dict())):
            self._assert_init(leaves, mock_generate, kwargs)

        # failure
        for leaves in (None, 73, [], [1, 2]):
            with self.assertRaises(ValueError):
                self._assert_init(leaves, mock_generate, {})
                assert not mock_generate.called

    def test_generate(self):
        """Test MerkleTree._generate()"""
        self.tree._make_row = mock.MagicMock()
        self.tree._make_row.return_value = ['a', 'b']

        self.tree._generate()

        make_row_calls = [mock.call(1), mock.call(0)]
        self.tree._make_row.assert_has_calls(make_row_calls)

    @mock.patch.object(ShardManager, 'hash', return_value='7')
    def test_make_row(self, mock_hash):
        """Test MerkleTree._make_row()"""

        self.tree._rows = [[],
                           [],
                           ['a', 'b', 'c', 'd']]

        row = self.tree._make_row(1)

        calls = [mock.call('ab'),
                 mock.call('cd')]
        mock_hash.assert_has_calls(calls)
        self.assertEqual(row, ['7', '7'])

    def test_property_depth(self):
        """Test depth property."""

        self.tree._leaves = mock.MagicMock()
        self.tree._leaves.__len__.return_value = 8

        depth = self.tree.depth

        self.assertEqual(self.tree.leaves.__len__.call_count, 4)
        self.assertEqual(2 ** depth, 8)

    def test_get_root(self):
        """Test MerkleTree.get_root"""
        root = self.tree.get_root()

        self.assertEqual(root, self.tree._rows[0][0])

    def test_get_level(self):
        """Test MerkleTree.get_level"""
        level = self.tree.get_level(1)

        self.assertEqual(level, self.tree._rows[1])

    def test_node_output(self):
        """Test MerkleTree with output from the node client"""
        leaves = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
        depth = 3
        rows = [['2adf050f14bf6324bfd41577d0dc08e2e49766fa'],
                ['39c096fc1b11e77f4347cfdb45ba9b03c0ad95d9',
                 '47121e7ec10e7653f1262b1d3abb6f9a71b3de8b'],
                ['e4973182d0c331ce8b083ffa2b28c8b4fc0f1d93',
                 'c91b9f3b2937035cc07d3fcd258d7d8a1f0c4d3c',
                 '8182daac9a266aa39328b835726f80a34835027d',
                 '222025114b2d1374b4a354d1b4452f648c9b481d'],
                ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']]
        count = 7

        tree = MerkleTree(leaves)

        self.assertEqual(tree.leaves, leaves)
        self.assertEqual(tree.depth, depth)
        self.assertEqual(tree._rows, rows)
        self.assertEqual(tree.count, count)


class TokenTestCase(AbstractTestCase):
    """Test case for the Token class."""

    def test_init(self):
        """Test Token.__init__()."""

        kwargs = dict(
            token=None,
            bucket='bucket_id',
            operation='operation',
            expires='2016-10-13T04:23:48.183Z',
            encryptionKey='key_id',
        )

        token = Token(**kwargs)

        assert token.id == kwargs['token']
        assert token.bucket == Bucket(id=kwargs['bucket'])
        assert token.operation == kwargs['operation']
        assert token.expires == datetime.fromtimestamp(
            strict_rfc3339.rfc3339_to_timestamp(kwargs['expires']))
        assert token.encryptionKey == kwargs['encryptionKey']



# -*- coding: utf-8 -*-
"""Storj API integration tests."""

import os

from datetime import datetime
from storj import http

from .. import AbstractTestCase


#: name of the environment variable that contains the Storj account email.
ENV_STORJ_EMAIL = 'STORJ_EMAIL'
#: name of the environment variable that contains the Storj account password.
ENV_STORJ_PASSWORD = 'STORJ_PASSWORD'


class Integration(AbstractTestCase):
    """Base class for integration tests.

    Attributes:
        email (str): Storj account email.
        password (str): Storj account password.
        client (:py:class:`storj.http.Client`): HTTP client.
        test_id (str): unique identifier to be appended to temporary resources.
    """

    def __init__(self, methodName):
        super(Integration, self).__init__(methodName)

        try:
            self.email = os.environ[ENV_STORJ_EMAIL]
            self.password = os.environ[ENV_STORJ_PASSWORD]
        except KeyError as e:
            self.logger.error(e)
            msg = (
                'To run integration tests you need to define '
                'the following environment variables: %s, %s.'
            )
            err_msg = msg % (
                ENV_STORJ_EMAIL, ENV_STORJ_PASSWORD
            )
            self.logger.error(err_msg)
            self.fail(err_msg)

    def setUp(self):
        super(Integration, self).setUp()
        self.client = http.Client(self.email, self.password)
        self.test_id = datetime.utcnow().strftime('%Y%m%d_%H%M%S')
